;; "blkio.scm" WB-tree File Based Associative String Data Base System.
;; Copyright (C) 1991, 1992, 1993, 2000, 2003 Free Software Foundation, Inc.
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU Lesser General Public License as
;; published by the Free Software Foundation, either version 3 of the
;; License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; Lesser General Public License for more details.
;;
;; You should have received a copy of the GNU Lesser General Public
;; License along with this program.  If not, see
;; <http://www.gnu.org/licenses/>.

;;; routines in this file return pass/fail (#t/#f)

(require 'i/o-extensions)

(pragma.c "#include \"wbsys.h\""
	  "#ifdef unix"
	  "# include <sys/types.h>"
	  "# include <sys/stat.h>"
	  "# include <fcntl.h>"
	  "# include <unistd.h>"
	  "#endif")

(java:import "Stats" "Wbdefs" "Wbsys" "Blk" "BooleanMethods")

(define io-diag #f)

(define (blkio:init!) #f)

(define (blkio:create-file name bsiz)
  (open-output-file name))

(define (blkio:open-modify-file name bsiz)
  (open-io-file name))

(define (blkio:open-read-only-file name bsiz)
  (open-input-file name))

(define (blkio:file-close file) (close-port file))

(define (blkio:read file blk bsiz blknum)
  (file-position file (* bsiz blknum))
  (cond ((= bsiz (subbytes-read! blk 0 bsiz file))
	 (if io-diag (dprintf "rd:%*s%c %10ld\n"
			      (+ 1 (- (BLK:LEVEL blk) LEAF)) ""
			      (BLK:TYP blk) blknum))
	 (if (BLK:TYP? blk FRL-TYP)
	     (set! read-fl-ct (+ read-fl-ct 1))
	     (set! read-ct (+ read-ct 1)))
	 #t)
	(else (edprintf " couldn't read blk %ld\n" blknum)
	      #f)))

(define (blkio:write file blk bsiz blknum)
  (file-position file (* bsiz blknum))
  (cond ((= bsiz (subbytes-write blk 0 bsiz file))
	 (if io-diag (dprintf "wr:%*s%c %10ld\n"
			      (+ 1 (- (BLK:LEVEL blk) LEAF)) ""
			      (BLK:TYP blk) blknum))
	 (if (BLK:TYP? blk FRL-TYP)
	     (set! write-fl-ct (+ write-fl-ct 1))
	     (set! write-ct (+ write-ct 1)))
	 #t)
	(else (edprintf " couldn't write blk %ld\n" blknum)
	      #f)))

(define (blkio:file-extend file blk bsiz blknum)
  (file-position file (* bsiz blknum))
  (cond ((= bsiz (subbytes-write blk 0 bsiz file))
	 (if io-diag (dprintf "Extending file %d blks\n"
			      (+ (quotient FLC-LEN 2) 1)))
	 #t)
	(else (edprintf " couldn't extend file %ld\n" blknum)
	      #f)))
