/* "Blk.cs" WB-tree File Based Associative String Data Base System.
 * Copyright (C) 1991, 1992, 1993, 2000, 2003 Free Software Foundation, Inc.
 * Copyright 2007 Clear Methods, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

/*  BLK parameters */
/*  The IDs are 4 byte numbers identifying this block, the root of */
/*  this tree, and the next in the chain. */
public static readonly int blkIdPos = 0;

public static readonly int blkTopIdPos = 4;

public static readonly int blkNxtIdPos = 8;

public static readonly int blkTimePos = 0xc;
/*  blk-end-pos is position (stored in 2 bytes) of first free byte */

public static readonly int blkEndPos = 0x10;

public static readonly int blkLevelPos = 0x12;

public static readonly int blkTypPos = 0x13;

public static readonly int blkDataStart = 0x14;

public static int blk_Id(byte[] blk)
{
	return str2long(blk, blkIdPos);
}

public static int  blk_TopId(byte[] blk)
{
	return str2long(blk, blkTopIdPos);
}

public static int  blk_NxtId(byte[] blk)
{
	return str2long(blk, blkNxtIdPos);
}

public static int  blk_Time(byte[] blk)
{
	return str2long(blk, blkTimePos);
}

public static int  blk_End(byte[] blk)
{
	return str2short(blk, blkEndPos);
}

public static int  blk_Level(byte[] b)
{
	return (b[blkLevelPos]);
}

public static int  blk_Typ(byte[] b)
{
	return (b[blkTypPos]);
}

public static bool  blk_Typ_P(byte[] b,int typ)
{
	return ((b[blkTypPos]))==(typ);
}

public static void  blk_SetId(byte[] blk,int id)
{
	long2str(blk, blkIdPos, id);
}

public static void  blk_SetTopId(byte[] blk,int id)
{
	long2str(blk, blkTopIdPos, id);
}

public static void  blk_SetNxtId(byte[] blk,int id)
{
	long2str(blk, blkNxtIdPos, id);
}

public static void  blk_SetTime(byte[] blk,int time)
{
	long2str(blk, blkTimePos, time);
}

public static void  blk_SetEnd(byte[] blk,int pos)
{
	short2str(blk, blkEndPos, pos);
}

public static void  blk_SetLevel(byte[] b,int level)
{
	//b[blkTypPos] = -128 + typ;
	b[blkLevelPos] = (byte) (0xFF & level);
}

public static void  blk_SetTyp(byte[] b,int typ)
{
	//b[blkTypPos] = -128+typ;
	b[blkTypPos] = (byte) (0xFF & typ);
}


/* "Db.cs" WB-tree File Based Associative String Data Base System. */

public static bool wb_Err_P(int x)
{
    return 0 > x && x >= maxerr;
}

public static int flushEnts(int trynum, int flushnum)
{
    return entsFlush(trynum, flushnum);
}

public static Han createBt(int seg, int typ, int wcb)
{
    Han aHan = hanMakeHan();
    int ans = btCreate(seg, typ, aHan, wcb);
    if (err_P(ans))
	return null;
    else return aHan;
}

public static Han openBt(int seg, int blknum, int wcb)
{
    Han aHan = hanMakeHan();
    int ans = btOpen(seg, blknum, aHan, wcb);
    if (err_P(ans))
	return null;
    else return aHan;
}

public static void closeBt(Han han)
{
    btClose(han);
}

public static Han createDb(int seg, char typ, String nameStr)
{
    return createDb(seg, (int) Char.GetNumericValue(typ), nameStr);
}

public static Han createDb(int seg,int typ,String nameStr)
{
    byte[] tmpStr = new byte[0x100];
    Han aHan = createBt(seg, typ, 0);
    Han dHan = openBt(seg, 1, (wcbSap + wcbSar));
    if (!(a2b(aHan)) && !(a2b(dHan))) return null;
    else {
	long2str(tmpStr, 1, han_Id(aHan));
	tmpStr[0] = 4;
	//should the name of the db too be UTF-8 encoded?
	btPut(dHan, stringToBytes(nameStr), nameStr.Length, tmpStr, 5);
	closeBt(dHan);
	return aHan;
    }
}

public static Han openDb(int seg, String nameStr)
{
    byte[] tmpStr = new byte[0x100];
    Han dHan = openBt(seg, 1, (wcbSap + wcbSar));
    if (dHan == null) return null;
    else {
	int tlen = btGet(dHan, stringToBytes(nameStr), nameStr.Length, tmpStr);
	closeBt(dHan);
	if (err_P(tlen)) return null;
	if(tlen == 5)
	    return openBt(seg, str2long(tmpStr, 1), 0);
	else return null;
    }
}

/*  Del removes key-str and value.  returns #t if found, #f if not. */
//bt-rem!
public static bool bt_Del(Han han, byte[] keyStr)
{
    return success_P(btRem(han, keyStr, keyStr.Length, null));
}

// bt-rem*
public static int bt_Delete(Han han, byte[] keyStr, byte[] key2Str)
{
    byte []tmpstr = new byte[0x100];
    substringMove(keyStr, 0, keyStr.Length, tmpstr, 0);
    return btRemRange(han, tmpstr, !(a2b(keyStr.Length))
		      ?startOfChain
		      :keyStr.Length, key2Str, !(a2b(key2Str.Length))
		      ?endOfChain
		      :key2Str.Length);
}

/*  rem removes key-str and value.  returns value. */
public static byte[] bt_Rem(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];
    int tlen = btRem(han, keyStr, keyStr.Length, tmpStr);
    if(err_P(tlen))
	return null;
    else
	return subArray(tmpStr, tlen);
    //return new String(tmpStr, 0, tlen);
}

/*  PUT adds an key-str value pair to the database whose root is blk */
//bt-put!
public static void bt_Put(Han han, byte[] keyStr, byte[] valStr)
{
    btPut(han, keyStr, keyStr.Length, valStr, valStr.Length);
}

//bt-put
public static bool bt_Insert(Han han, byte[] keyStr, byte[] valStr)
{
    return success_P(btWrite(han,
			     keyStr, keyStr.Length,
			     valStr, valStr.Length));
}

/*  get returns a string of the value or #f */
public static byte[] bt_Get(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];

    int tlen = btGet(han, keyStr, keyStr.Length, tmpStr);
    if (err_P(tlen))
	return null;
    else
	return subArray(tmpStr, tlen);
    //return new String(tmpStr, 0, tlen);
}

/*  next returns a string of the next key-str or #f if at end. */
/*  (bt:next blk #f) or (bt:next blk "") returns the first key-str. */
/*  to make BLINK happy I'm passing length 0 instead of START-OF-CHAIN */
public static byte[] bt_Next(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];
    int tlen = (a2b(keyStr) && (keyStr.Length>0))
	?btNext(han, keyStr, keyStr.Length, tmpStr)
	:btNext(han, noByts, startOfChain, tmpStr);
    if (err_P(tlen))
	return null;
    else
	return subArray(tmpStr, tlen);
    //return new String(tmpStr, 0, tlen);
}

/*  prev returns a string of the previous key-str or #f if at end. */
/*  (bt:prev blk #f) or (bt:prev blk "") returns the last key-str. */
public static byte[] bt_Prev(Han han, byte[] keyStr)
{
    byte []tmpStr = new byte[0x100];
    int tlen = (a2b(keyStr)
		&& a2b(keyStr.Length>0))
	?btPrev(han, keyStr, keyStr.Length, tmpStr)
	:btPrev(han, noByts, endOfChain, tmpStr);

    if (err_P(tlen))
	return null;
    else
	return subArray(tmpStr, tlen);
    //return new String(tmpStr, 0, tlen);
}

/* Wrapper functions for db access that take strings instead
   of byte arrays */
// public static int bt_Del(Han han, String keyStr)
// {
//     return bt_Del(han, stringToBytes(keyStr));
// }

public static void bt_Delete(Han han, String keyStr, String key2Str)
{
    bt_Delete(han, stringToBytes(keyStr), stringToBytes(key2Str));
}

public static String bt_Get(Han han, String keyStr)
{
    byte[] byts = bt_Get(han, stringToBytes(keyStr));
    return bytesToString(byts);
}

public static String bt_Next(Han han, String keyStr)
{
    byte[] byts = bt_Next(han, stringToBytes(keyStr));
    return  bytesToString(byts);
}


public static String bt_Prev(Han han, String keyStr)
{
    byte[] byts =  bt_Prev(han, stringToBytes(keyStr));
    return bytesToString(byts);
}

public static void bt_Put(Han han, String keyStr, String key2Str)
{
    bt_Put(han, stringToBytes(keyStr), stringToBytes(key2Str));
}

public static String bt_Rem(Han han, String keyStr)
{
    byte[] byts =  bt_Rem(han, stringToBytes(keyStr));
    return bytesToString(byts);
}

public static byte[] int2bytes(int i)
{
    byte[] B = new byte[4];
    long2str(B, 0, i);
    return B;
}

/* "Ent.cs" WB-tree File Based Associative String Data Base System. */

public class Ent{
	public Ent NEXT;
	public int ID;
	public byte[] BLK;
	public bool DTY;
	public int TAG, AGE,  PUS, ACC, REF, SEG;
}

public static Ent makeEntry(){
	return new Ent();
}

public static Ent ent_MakeEnt(int tag){
	//System.out.println("blkSize = " + blkSize);
	Ent ent = new Ent();
	ent.TAG = tag;
	ent.SEG = -1;
	ent.ID  = -1;
	ent.BLK = new byte[blkSize];
	return ent;
}

//The getter methods in C take struct entry as a parameter.
// Here we use object methods, thus the function ent_tag(ent)
//in C appears as ent.entTag() in Java.
public static int  ent_Tag(Ent entry){
	return entry.TAG;
}

public  static Ent  ent_Next(Ent entry){
	return entry.NEXT;
}

public static int  ent_Seg(Ent entry){
	return entry.SEG;
}

public static int  ent_Id(Ent entry){
	return entry.ID;
}

public static byte[]  ent_Blk(Ent entry){
	return entry.BLK;
}

public static int  ent_Age(Ent entry){
	return entry.AGE;
}

public static bool  ent_Dty_P(Ent entry){
	return entry.DTY;
}

public static int  ent_Pus(Ent entry){
	return entry.PUS;
}

public static int  ent_Acc(Ent entry){
	return entry.ACC;
}

public static int  ent_Ref(Ent entry){
	return entry.REF;
}
//As with getter methods, we do not take  ry object as parameter,
//but invoke methods on the objects.

public static void  ent_SetTag(Ent entry, int tag){
	entry.TAG = tag;
}

public static void  ent_SetNext(Ent entry, Ent next){
	entry.NEXT = next;
}

public static void  ent_SetSeg(Ent entry, int seg){
	entry.SEG = seg;
}

public static void  ent_SetId(Ent entry, int num){
	entry.ID = num;
}

public static void  ent_SetAge(Ent entry, int age){
	entry.AGE = age;
}

public static void  ent_SetDty(Ent entry, Object A){//for handling nulls.
	entry.DTY = ((A==null)? false:true);
}

public static void  ent_SetDty(Ent entry, bool dty){
	entry.DTY = dty;
}

public static void  ent_SetPus(Ent entry, int pus){
	entry.PUS = pus;
}

public static void  ent_SetAcc(Ent entry, int acc){
	entry.ACC = acc;
}

public static void  ent_SetRef(Ent entry, int Ref){
	entry.REF = Ref;
}

/* "Han.cs" WB-tree File Based Associative String Data Base System. */

public class Han{
	public int SEG, TYP;
	public int ID, LAST;
	public int WCB, SPARE;
}

public static Han hanMakeHan(){
	return new Han();
}

//The getter & setter methods follow the same syntax translations as
//the corresponding methods in Entry.
public static int han_Id(Han han){
	return han.ID;
}

public static int han_Seg(Han han){
	return han.SEG;
}

public static int han_Typ(Han han){
	return han.TYP;
}

public static int han_Last(Han han){
	return han.LAST;
}

public static int han_Wcb(Han han){
	return han.WCB;
}

//setter
public static void han_SetNum(Han han, int num){
	han.ID = num;
}

public static void han_SetSeg(Han han, int seg){
	han.SEG = seg;
}

public static void han_SetTyp(Han han, int typ){
	han.TYP = typ;
}

public static void han_SetLast(Han han, int last){
	han.LAST = last;
}

public static void han_SetWcb(Han han, int wcb){
	han.WCB = wcb;
}

/* "Lck.cs" WB-tree File Based Associative String Data Base System. */

public class Lck {
    public Lck next;
    public int flg;
    public int name;
}

public static Lck lastLck;

public static Lck makeLck(int name) {
    Lck lk = new Lck();
    lk.next = lastLck;
    Interlocked.Exchange(ref lk.flg, 0);
    lk.name = name;
    lastLck = lk;
    return lk;
}

public static bool tryLck_P(Lck lk) {
    if (Interlocked.Decrement(ref lk.flg) > 0) return false;
    return true;
}

public static void lck(Lck lk) {
    if (Interlocked.Decrement(ref lk.flg) > 0)
	dprintf(">>>>ERROR<<<< spinning on lck " + lk.name);
}

public static void unlck(Lck lk) {
    if (lk.flg == 0)
	dprintf(">>>>ERROR<<<< unlcking unlck "+ lk.name);
    Interlocked.Exchange(ref lk.flg, 0);
}

public static void checkLcks() {
    Lck lck = lastLck;
    while(lck!=null) {
	if(lck.flg != 0) {
	    if(lck.name < 0)
		switch(lck.name) {
		case -1:
		    dprintf(">>>>ERROR<<<< free-ent-lck left lcked\n");
		    break;
		case -2:
		    dprintf(">>>>ERROR<<<< flush-buk-lck left lcked\n");
		    break;
		default:
		    dprintf(">>>>ERROR<<<< unknown lck left lcked\n");
		    break;
		}
	    else
		dprintf(">>>>ERROR<<<< lck " + lck.name + "  left lcked\n");
	    Interlocked.Exchange(ref lck.flg, 0);
	}
	lck = lck.next;
    }
}

/* "Pkt.cs" WB-tree File Based Associative String Data Base System. */

/*  Routines for finding the appropriate BLK for an operation. */
/*  PACKETs used to return multiple values from chain-find. */
/*  and various other operations */

public static readonly int pktSize = 6;


public static int[] makePkt()
{
	return new int[pktSize];
}

public static int pkt_MatchType(int[] p)
{
	return p[0];
}

public static int  pkt_MatchPos(int[] p)
{
	return p[1];
}

public static int  pkt_KeyPos(int[] p)
{
	return p[2];
}

public static int  pkt_PrevMatchPos(int[] p)
{
	return p[3];
}

public static int  pkt_BlkToCache(int[] p)
{
	return p[4];
}

public static int  pkt_SuccessCode(int[] p)
{
	return p[5];
}

public static void  pkt_SetMatchType(int[] p,int v)
{
	p[0] = v;
}

public static void  pkt_SetMatchPos(int[] p,int v)
{
	p[1] = v;
}

public static void  pkt_SetKeyPos(int[] p,int v)
{
	p[2] = v;
}

public static void  pkt_SetPrevMatchPos(int[] p,int v)
{
	p[3] = v;
}

public static void  pkt_SetBlkToCache(int[] p,int v)
{
	p[4] = v;
}

public static void  pkt_SetSuccessCode(int[] p,int v)
{
	p[5] = v;
}

//Aliases
public static int pkt_SkeyCount(int[] p){
	return pkt_MatchPos(p);
}

public static void pkt_SetSkeyCount(int[] p, int v){
	pkt_SetMatchPos(p, v);
}

public static int pkt_SkeyLen (int[] p){
	return pkt_KeyPos(p);
}

public static void pkt_SetSkeyLen (int[]p, int v){
	pkt_SetKeyPos(p, v);
}

public static void pkt_Pack(int[] p,int type,int bPos,int kPos,int pPos)
{
	pkt_SetMatchType(p, type);
	pkt_SetMatchPos(p, bPos);
	pkt_SetKeyPos(p, kPos);
	pkt_SetPrevMatchPos(p, pPos);
}

/* "Seg.cs" WB-tree File Based Associative String Data Base System. */

public class Seg {
    public FileStream PORT;
    public int BSIZ;
    public int USED;
    public String STRN;
    public Han RT_HAN, FL_HAN;
    public Lck FLCK ;
    public Lck FFCK ;
    public int FLC_LEN;
    public int []FLC;

    public Seg(){
	this.RT_HAN = hanMakeHan();
	this.FL_HAN = hanMakeHan();
	this.FLCK = makeLck(0);
	this.FFCK = makeLck(0);
    }

    public static Seg[] segmentTable = new Seg[numSegs];

    //In Java, the static block gets executed at load time.
    //In C#, the static block gets executed when,
    //1. A static member of the class gets accessed, or
    //2. A instance of class is created.
    static Seg(){
	//Console.WriteLine("In Seg's static Initializer");
	//Console.WriteLine("segmentTable.Length = " + segmentTable.Length);
	for(int i = 0; i < numSegs; i++)
	    segmentTable[i] = new Seg();
	//Console.WriteLine("segmentTable.Length = " + segmentTable.Length);
	//Console.WriteLine("numSegs = " + numSegs);
    }
}

/*  The SEG (segment) methods*/
public static FileStream seg_Port(int seg)
{
    return  segdTab[seg].PORT;
}

public static int  seg_Bsiz(int seg)
{
    return  segdTab[seg].BSIZ;
}

public static int  seg_Used(int seg)
{
    return  segdTab[seg].USED;
}

public static String  seg_Str(int seg)
{
    return  segdTab[seg].STRN;
}

public static Han seg_RtHan(int seg)
{
    return  segdTab[seg].RT_HAN;
}

public static Han seg_FlHan(int seg)
{
    return  segdTab[seg].FL_HAN;
}

public static Lck  seg_Lck(int seg)
{
    return  segdTab[seg].FLCK;
}

public static Lck  seg_Fck(int seg)
{
    return  segdTab[seg].FFCK;
}

public static int  seg_FlcLen(int seg)
{
    return  segdTab[seg].FLC_LEN;
}

public static int[]  seg_Flc(int seg)
{
    return  segdTab[seg].FLC;
}

public static void  seg_SetPort(int seg,FileStream port)
{
    segdTab[seg].PORT = port;
}

public static void  seg_SetBsiz(int seg,int bsiz)
{
    segdTab[seg].BSIZ =  bsiz;
}

public static void  seg_SetUsed(int seg,int used)
{
    segdTab[seg].USED =  used;
}

public static void  seg_SetStr(int seg,String str)
{
    segdTab[seg].STRN =  str;
}

public static void  seg_SetFlcLen(int seg,int flcLen)
{
    segdTab[seg].FLC_LEN = flcLen;
}

public static void  seg_SetFlc(int seg,int[] flc)
{
    segdTab[seg].FLC = flc;
}

public static Seg newSeg()
{
    return new Seg();
}

public static Seg[] segdTab = Seg.segmentTable;
