/*  "handle.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000, 2007 Free Software Foundation, Inc. */
/*  */
/*  This program is free software: you can redistribute it and/or modify */
/*  it under the terms of the GNU Lesser General Public License as */
/*  published by the Free Software Foundation, either version 3 of the */
/*  License, or (at your option) any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, but */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*  Lesser General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU Lesser General Public */
/*  License along with this program.  If not, see */
/*  <http://www.gnu.org/licenses/>. */

/*  routines in this file return success codes */


#include "wbsys.h"


int clever_cache_enable = !0;

/* Note: most of the data-manipulating commands here */
/* can return @var{notpres}, with the followng meanings: */
/* @cindex notpres */
/*  */
/* @multitable @columnfractions .2 .8 */
/* @item @code{bt-get} */
/* @tab no such key. */
/* @item @code{bt-next} */
/* @tab no NEXT key (ie, key given was LAST key). */
/* @item @code{bt-prev} */
/* @tab no PREV key (ie, key given was FIRST key). */
/* @item @code{bt-rem} */
/* @tab KEY was not found. */
/* @c @item @code{bt-rem-range} */
/* @c @tab ?? */
/* @item @code{bt-put} */
/* @tab NOT USED (could be symmetric w/WRITE). */
/* @item @code{bt-write} */
/* @tab key WAS found, so no write done. */
/* @end multitable */
/* @var{key_str} is a string of length @var{k_len}.  @code{bt_get} stores into the string @var{ans_str} the */
/* value associated with @var{key_str} in tree @var{han}.  @code{bt_get} returns the length of the */
/* string stored into @var{ans_str} or an error code. */

int bt_get(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  ent = chain_find_ent(han, accread, key_str, k_len, pkt);
  if (!(ent))
    {
      get_fct = 0x1L+(get_fct);
      return unkerr;
    }
  else if ((pkt_match_type(pkt))!=(match))
    {
      get_ct = 0x1L+(get_ct);
      release_ent(ent, accread);
      return notpres;
    }
  else {
    int alen = get_this_val(ent_blk(ent), pkt_match_pos(pkt), ans_str);
    get_ct = 0x1L+(get_ct);
    release_ent(ent, accread);
    return alen;
  }
}

/* @var{key_str} is a string of length @var{k_len}.  @code{bt_next} stores into the string @var{ans_str} the */
/* next key after @var{key_str} in tree @var{han}.  @code{bt_next} returns the length of the string */
/* stored into @var{ans_str} or an error code. */

int bt_next(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  ent = chain_find_ent(han, accread, key_str, k_len, pkt);
  if (!(ent))
    {
      next_fct = 0x1L+(next_fct);
      return unkerr;
    }
  else {
    next_ct = 0x1L+(next_ct);
    {
      int res = chain_next(ent, key_str, k_len, ans_str, pkt);
      if (clever_cache_enable)
	han_set_last(han, pkt_blk_to_cache(pkt));
      return res;
    }
  }
}

/* @var{key_str} is a string of length @var{k_len}.  @code{bt_prev} stores into the string @var{ans_str} the */
/* last key before @var{key_str} in tree @var{han}.  @code{bt_prev} returns the length of the string */
/* stored into @var{ans_str} or an error code. */

int bt_prev(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  ent = chain_find_prev_ent(han, accread, key_str, k_len, pkt);
  if (ent)
    ent = prev_prev_k_ent(ent, key_str, k_len, leaf, pkt);
  if (!(ent))
    {
      prev_fct = 0x1L+(prev_fct);
      return unkerr;
    }
  else {
    prev_ct = 0x1L+(prev_ct);
    if (!(pkt_match_pos(pkt)))
      {
	release_ent(ent, accread);
	return notpres;
      }
    else {
      int k_len2 = recon_this_key(ent_blk(ent), pkt_match_pos(pkt), ans_str, 0, 0x100);
      han_set_last(han, ent_id(ent));
      release_ent(ent, accread);
      return k_len2;
    }
  }
}


/* @var{key_str} is a string of length @var{k_len}.  @code{bt_rem} stores into the string @var{ans_str} the */
/* value associated with @var{key_str} in tree @var{han}; then removes that association */
/* from tree @var{han}.  @code{bt_rem} returns the length of the string stored into @var{ans_str} or */
/* an error code. */
/*  */
/* If @var{ans_str} is 0, @code{bt_rem} removes the @var{key_str} association from tree @var{han} and returns */
/* @var{SUCCESS} if successful; an error code if not. */

int bt_rem(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  if ((k_len)<0)
    {
      dprintf((diagout, ">>>>ERROR<<<< %s: bad length string %d\n", "bt_rem", k_len));
      return argerr;
    }
  else {
    ent = chain_find_ent(han, accwrite, key_str, k_len, pkt);
    if (ent)
      {
	rem_ct = 0x1L+(rem_ct);
	{
	  int ans = del_chain_rem(ent, key_str, k_len, ans_str, pkt, han_wcb(han));
	  release_ent(ent, accwrite);
	  return ans;
	}
      }
    else {
      rem_fct = 0x1L+(rem_fct);
      return unkerr;
    }
  }
}

/*  To make possible bounded-time operation, @0 will purge at most */
/*  BLK-LIMIT blocks at a time; passing BLK-LIMIT of -1 imposes no */
/*  limit. */

/* @var{key_str} must be a maximum-length (256 byte) string containing a key */
/* @var{k_len} bytes long.  @var{key2_str} is a string of length @var{k2_len}. */
/*  */
/* @code{bt_rem_range} removes [@var{key_str} @dots{} @var{key2_str}) and their values.  If @var{key2_str} <= @var{key_str} no */
/* deletion will occur (even if @var{key_str} is found).  @code{bt_rem_range} returns SUCCESS if */
/* the operation is complete, an error status code if not. */

int bt_rem_range(han, key_str, k_len, key2_str, k2_len)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *key2_str;
     int k2_len;
{
  int respkt[pkt_size];
  pkt_set_skey_count(respkt, 0);
  return bt_scan(han, rem_scan, key_str, k_len, key2_str, k2_len, 0, 0, respkt,  -1);
}

/* @var{key_str} is a string of length @var{k_len}.  @var{val_str} is a string of length @var{v_len}.  @code{bt_put} */
/* makes the value associated with @var{key_str} be @var{val_str} in tree @var{han}.  @code{bt_put} returns a */
/* status code for the operation. */

int bt_put(han, key_str, k_len, val_str, v_len)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *val_str;
     int v_len;
{
  ENTRY *ent = 0;
  int pkt[pkt_size];
  if (((v_len)>0xff)
      || ((k_len)>0xff)
      || ((k_len)<0))
    return argerr;
  else {
    ent = chain_find_ent(han, accwrite, key_str, k_len, pkt);
    if (ent)
      {
	int res_P = chain_put(ent, key_str, k_len, val_str, v_len, pkt, 0, han_wcb(han));
	if (res_P)
	  {
	    if (clever_cache_enable)
	      han_set_last(han, pkt_blk_to_cache(pkt));
	    put_ct = 0x1L+(put_ct);
	    return success;
	  }
	else {
	  put_fct = 0x1L+(put_fct);
	  return unkerr;
	}
      }
    else return unkerr;
  }
}

/* @var{key_str} is a string of length @var{k_len}.  @var{val_str} is a string of length @var{v_len}.  If */
/* @var{han} currently contains an association for @var{key_str}, then @code{bt_write} does not modify */
/* the tree and returns the @var{notpres} status code. */
/*  */
/* Otherwise, @code{bt_write} makes the value associated with @var{key_str} be @var{val_str} in tree @var{han}. */
/* @code{bt_write} returns a status code for the operation. */

int bt_write(han, key_str, k_len, val_str, v_len)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *val_str;
     int v_len;
{
  ENTRY *ent = 0;
  int pkt[pkt_size];
  if (((v_len)>0xff)
      || ((k_len)>0xff)
      || ((k_len)<0))
    return argerr;
  else {
    ent = chain_find_ent(han, accwrite, key_str, k_len, pkt);
    if (ent)
      if ((pkt_match_type(pkt))==(match))
	{
	  release_ent(ent, accwrite);
	  return notpres;
	}
      else {
	int res_P = chain_put(ent, key_str, k_len, val_str, v_len, pkt, 0, han_wcb(han));
	if (res_P)
	  {
	    if (clever_cache_enable)
	      han_set_last(han, pkt_blk_to_cache(pkt));
	    put_ct = 0x1L+(put_ct);
	    return success;
	  }
	else {
	  put_fct = 0x1L+(put_fct);
	  return unkerr;
	}
      }
    else return unkerr;
  }
}
	
