/*  "scan.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000 Free Software Foundation, Inc. */
/*  */
/*  This program is free software: you can redistribute it and/or modify */
/*  it under the terms of the GNU Lesser General Public License as */
/*  published by the Free Software Foundation, either version 3 of the */
/*  License, or (at your option) any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, but */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*  Lesser General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU Lesser General Public */
/*  License along with this program.  If not, see */
/*  <http://www.gnu.org/licenses/>. */



package wb;

import static wb.Ents.*;
import static wb.Stats.*;
import static wb.Blink.*;
import static wb.Del.*;
import static wb.Handle.*;
import static wb.Wbdefs.*;
import static wb.Wbsys.*;
import static wb.Blk.*;
import static wb.BooleanMethods.*;
import static wb.Pkt.*;
import static wb.Seg.*;
import static wb.Han.*;
import static wb.Ent.*;
public class Scan {

/* @code{btScan} scans all keys in the range [@var{kstr1}..@var{kstr2}), */
/* performing one of several functions: */
/*  */
/* @multitable @columnfractions .2 .15 .66 */
/* @item @var{operation} */
/* @tab @var{func} */
/* @tab RESULT */
/* @item COUNT-SCAN */
/* @tab NIL */
/* @tab counts all keys in range */
/* @item COUNT-SCAN */
/* @tab given */
/* @tab counts all keys in range satisfying @var{func} */
/* @item REM-SCAN */
/* @tab NIL */
/* @tab deletes all keys in range */
/* @item REM-SCAN */
/* @tab given */
/* @tab deletes all keys in range satisfying @var{func} */
/* @item MODIFY-SCAN */
/* @tab NIL */
/* @tab ARGERR */
/* @item MODIFY-SCAN */
/* @tab given */
/* @tab updates values for keys in range satisfying @var{func} */
/* @end multitable */
/*  */
/* @code{btScan} returns SUCCESS if scan completed; under any other result code */
/* the scan is resumable. The possible results are: */
/*  */
/* @table @asis */
/* @item NOTPRES */
/* meaning the @var{blkLimit} was exceeded; */
/* @item RETRYERR */
/* meaning @var{func} or delete got a RETRYERRR; */
/* @item TERMINATED */
/* meaning @var{func} asked to terminate the scan; */
/* @item <other error> */
/* means @var{func} or DELETE encountered this errror. */
/* @end table */
/*  */
/* Each block of data is scanned/deleted/modified in a single operation */
/* that is, the block is found and locked only once, and only written after */
/* all modifications are made. Tho only exception is that MODIFY-SCANs */
/* that increase the size of values  can cause block splits. Such cases */
/* are detected and converted to a PUT plus a NEXT. This has */
/* two consequences: data is written out each time a PUT occurs, */
/* and it is conceivable that @var{func} may be called more than once on the */
/* key value that caused the split if a RETRYERR occurs in the PUT. */
/* However, SCAN guarantees that only one modification will actually be */
/* made in this case (so that one can write INCREMENT-RANGE, for example). */
/*  */
/* @var{func} is passed pointers to (copies of) the key and value, */
/* plus one user argument: */
/* @example */
/* @var{func} (keystr klen vstr vlen extra_arg); */
/* @end example */
/* @var{func} is expected to return either: SUCCESS for DELETE/COUNT, */
/* NOTPRES/NOTDONE for SKIP (ie, DONT DELETE/COUNT), or */
/* any other code to terminate the scan resumably at the current point. */
/* For MODIFY-SCAN, if changing the value, the new value length is returned. */
/* Except for the case mentioned above, the caller can depend on @var{func} */
/* being called exactly once for each key value in the specified range, */
/* and only on those values. */
/*  */
/* If @var{kstr2} <= @var{kstr1}, then no scan will occur (even if @var{kstr1} is found). */
/* To make possible bounded-time operation @code{btScan} will */
/* access at most @var{blkLimit} blocks at a time; if you dont care, */
/* give it -1 for @var{blkLimit}. */
/*  */
/* The number of keys deleted/counted/modified is returned in the */
/* @code{skey-count} field of @var{respkt}; the key to resume at is returned in */
/* @var{kstr1} (@emph{which therefore needs to be 256 bytes long}); and the */
/* new key length is returned in the @code{skey-len} field of @var{respkt}. If */
/* returns SUCCESS, @code{skey-len} is zero.  NOTE that */
/* @code{skey-count} is cumulative, so the caller needs to initialize */
/* it to 0 when starting a new @code{btScan}. */
/*  */
/* @emph{WARNING:} when @code{btScan} returns other than SUCCESS, it modifies */
/* the @var{kstr1} string so that the string args are correctly set up for the */
/* next call (the returned value is the new length for @var{kstr1}). */
/* Therefore, @emph{@var{kstr1} must be a maximum-length string!} */

public static int btScan(Han han,int operation,byte []kstr1,int len1,byte []kstr2,int len2,java.lang.reflect.Method func,int []longTab,int []respkt,int blkLimit)
{
LbtScan:while (true) {
  {
    int []pkt = new int[pktSize];
    int []opkt = new int[pktSize];
    Ent ent = null;
    byte []vstr = new byte[0x100];
    int accmode = (operation)==(countScan)
      ?accread
      :accwrite;
    int result = success;
    if ((len1)< -2)
      {
	dprintf(">>>>ERROR<<<< "+("btScan")+": bad length string1 "+(len1)+"\n");
	return argerr;
      }
    else if ((len2)< -1)
      {
	dprintf(">>>>ERROR<<<< "+("btScan")+": bad length string2 "+(len2)+"\n");
	return argerr;
      }
    else if ((operation)==(modifyScan)
        && !(a2b(func)))
      {
	dprintf(">>>>ERROR<<<< "+("btScan")+": MODIFY-SCAN requires func be specified\n");
	return argerr;
      }
    else {
      ent = chainFindEnt(han, accmode, kstr1, len1, pkt);
      if (a2b(ent)
	  && a2b(blk_FindPos(ent_Blk(ent), kstr2, len2, opkt)))
	{
	  if ((operation)==(countScan))
	    {
	      Ent nent = allocateEnt();
	      entCopy(nent, ent);
	      releaseEnt(ent, accmode);
	      result = chainScan(nent, operation, pkt, opkt, kstr1, func, longTab, vstr, respkt, han_Wcb(han));
	      recycleEnt(nent);
	    }
	  else {
	    result = chainScan(ent, operation, pkt, opkt, kstr1, func, longTab, vstr, respkt, han_Wcb(han));
	    releaseEnt(ent, accmode);
	    if ((result)>0)
	      {
		result = btPut(han, kstr1, pkt_SkeyLen(respkt), vstr, result);
		if ((result)==(success))
		  {
		    pkt_SetSkeyCount(respkt, (pkt_SkeyCount(respkt))+1);
		    pkt_SetSkeyLen(respkt, incrementString(kstr1, pkt_SkeyLen(respkt), 0x100));
		    result = notpres;
		  }
	      }
	  }
	  if ((result)==(notpres)
	      && 0!=(blkLimit))
	    {
	      len1 = pkt_SkeyLen(respkt);
	      blkLimit = (blkLimit)-1;
	      continue LbtScan;
	    }
	  else return result;
	}
      else {
	if (a2b(ent))
	  releaseEnt(ent, accmode);
	remFct = 1+(remFct);
	return unkerr;
      }
    }
  }
}
}

/*  this function increments a string lexicographically */

public static int incrementString(byte []str,int len,int maxlen)
{
  if ((len)<(maxlen))
    {
      str[len] = (byte) (0xff&0);
      return (len)+1;
    }
  else {
    int oldval = (str[(len)-1] & 0xFF);
    str[(len)-1] = (byte) (0xff&(1+(oldval)));
    return len;
  }
}

/*  Each call to CHAIN-SCAN scans */
/*  all the keys within the specified range WITHIN block ENT. */
/*  If the scan actually reaches  the end of range, it sets SKEY-LEN=0 */
/*  and returns SUCCESS. If there's more to the range, */
/*  it sets KEY-STR to the key to continue deleting */
/*  from (ie, the split key of ENT), SKEY-LEN to its length, and */
/*  returns NOTPRES (NOTDONE). The caller must then call CHAIN-FIND */
/*  to find the START and END keys and call again. */

public static int chainScan(Ent ent,int operation,int []pkt,int []opkt,byte []keyStr,java.lang.reflect.Method func,int []longTab,byte []vstr,int []respkt,int wcb)
{
  {
    byte []blk = ent_Blk(ent);
    int result = success;
    if ((operation)==(remScan)
        && !(a2b(func))
        && (pkt_MatchPos(opkt))>(pkt_MatchPos(pkt))
        && (pkt_MatchPos(pkt))==(blkDataStart)
        && atSplitKeyPos_P(blk, pkt_MatchPos(opkt)))
      {
	{
	  int keyLen = reconThisKey(blk, pkt_MatchPos(opkt), keyStr, 0, 0x100);
	  substringMove(keyStr, 0, keyLen, blk, (blkDataStart)+2);
	  setFieldLen(blk, (blkDataStart)+1, keyLen);
	  blk_SetEnd(blk, (blkDataStart)+2+(keyLen));
	}
	pkt_SetSkeyCount(respkt, (pkt_SkeyCount(respkt))+1);
	remCt = 1+(remCt);
	ent_SetDty(ent, true);
	pkt_SetMatchPos(opkt, blkDataStart);
      }
    else {
      int oldct = pkt_SkeyCount(respkt);
      byte []ckstr = new byte[0x100];
      int clen = 0;
      if (a2b(func))
	clen = reconThisKey(blk, pkt_MatchPos(pkt), ckstr, 0, 0x100);
      pkt_SetMatchType(pkt, match);
      result = scanLoop(ent_Blk(ent), operation, pkt, opkt, func, longTab, respkt, ckstr, clen, vstr, seg_Bsiz(ent_Seg(ent)));
      if ((operation)!=(countScan)
	  && (pkt_SkeyCount(respkt))>(oldct))
	ent_SetDty(ent, true);
    }
    if ((operation)==(remScan)
        && blkEmpty_P(blk)
        && !(endOfChain_P(blk)))
      {
	del_DeleteBck(ent);
      }
    else if (ent_Dty_P(ent))
      if (((operation)==(remScan)
	      && (0 != ((wcbSar)&(wcb))
		     || (blk_Level(blk))>(leaf))
	  || (operation)==(modifyScan)
		 && 0 != ((wcbSap)&(wcb))))
	ents_EntWrite(ent);
    if ((result)!=(success))
      {
	pkt_SetSkeyLen(respkt, reconThisKey(blk, pkt_MatchPos(pkt), keyStr, 0, 0x100));
	return result;
      }
    else if ((pkt_MatchType(opkt))==(pastend)
        && !(endOfChain_P(blk)))
      {
	pkt_SetSkeyLen(respkt, reconThisKey(blk, pkt_MatchPos(pkt), keyStr, 0, 0x100));
	return notpres;
      }
    else {
      pkt_SetSkeyLen(respkt, 0);
      return success;
    }
  }
}

/*  SCAN-LOOP returns SUCCESS if it reaches the end of the range, */
/*  else an ERROR code if terminated before that point, either */
/*  by an error or by FUNC returning TERMINATED. */
/*  SCAN-LOOP returns a value>0 to signal the case of */
/*  a MODIFY that requires a block-split. That value is the */
/*  length of the new value (which must be >0 to have caused an */
/*  increase in block size). SCAN-LOOP NEVER returns NOTPRES. */
/*  Note that (PKT:MATCH-POS pkt) is always the current scan point. */


public static int scanLoop(byte []blk,int operation,int []pkt,int []opkt,java.lang.reflect.Method func,int []longTab,int []respkt,byte []ckstr,int clen,byte []vstr,int blksize)
{
LscanLoop:while (true) {
  if ((pkt_MatchPos(opkt))>(pkt_MatchPos(pkt)))
    {
      int oldBend = blk_End(blk);
      int nextPos = nextCnvpair(blk, pkt_MatchPos(pkt));
      int result = success;
      if (a2b(func))
	{
	  int vpos = nextField(blk, 1+(pkt_MatchPos(pkt)));
	  int vlen = fieldLen(blk, vpos);
	  substringMove(blk, (vpos)+1, (vpos)+(vlen)+1, vstr, 0);
	  result = intFunInvoke(func, null, new Object[] {ckstr, clen, vstr, vlen, longTab});
	}
      if ((result)>=(success))
	{
	  if ((operation)==(remScan))
	    {
	      blk_RemoveKeyAndVal(blk, pkt_MatchPos(pkt), blksize);
	      pkt_SetSkeyCount(respkt, (pkt_SkeyCount(respkt))+1);
	      remCt = 1+(remCt);
	      if ((pkt_MatchPos(opkt))==(nextPos))
		{
		  pkt_SetMatchPos(opkt, pkt_MatchPos(pkt));
		}
	      else pkt_SetMatchPos(opkt, (pkt_MatchPos(opkt))-((oldBend)-(blk_End(blk))));
	      nextPos = pkt_MatchPos(pkt);
	    }
	  else if ((operation)==(countScan))
	    {
	      pkt_SetSkeyCount(respkt, (pkt_SkeyCount(respkt))+1);
	      pkt_SetMatchPos(pkt, nextPos);
	    }
	  else if (a2b(changeExistingValue(blk, pkt_MatchPos(pkt), ckstr, clen, vstr, result, blksize)))
	    {
	      pkt_SetSkeyCount(respkt, (pkt_SkeyCount(respkt))+1);
	      nextPos = (nextPos)-((oldBend)-(blk_End(blk)));
	      pkt_SetMatchPos(opkt, (pkt_MatchPos(opkt))-((oldBend)-(blk_End(blk))));
	      pkt_SetMatchPos(pkt, nextPos);
	      result = success;
	    }
	  else dprintf(""+("scanLoop")+": hit modify special case\n");
	}
      else if ((result)==(notpres))
	pkt_SetMatchPos(pkt, nextPos);
      if (((result)==(success)
	  || (result)==(notpres)))
	{
	  if (a2b(func))
	    {
	      clen = (fieldLen(blk, nextPos))+(fieldLen(blk, 1+(nextPos)));
	      substringMove(blk, (nextPos)+2, (nextPos)+2+(fieldLen(blk, 1+(nextPos))), ckstr, fieldLen(blk, nextPos));
	    }
	  continue LscanLoop;
	}
      else return result;
    }
  else return success;
}
}

}
