/*  "segs.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000, 2007 Free Software Foundation, Inc. */
/*  */
/*  This program is free software: you can redistribute it and/or modify */
/*  it under the terms of the GNU Lesser General Public License as */
/*  published by the Free Software Foundation, either version 3 of the */
/*  License, or (at your option) any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, but */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*  Lesser General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU Lesser General Public */
/*  License along with this program.  If not, see */
/*  <http://www.gnu.org/licenses/>. */

/*  routines in this file return success codes */



package wb;

import static wb.Ents.*;
import static wb.Blkio.*;
import static wb.Blink.*;
import static wb.Wbdefs.*;
import static wb.Handle.*;
import static wb.Wbsys.*;
import static wb.Blk.*;
import static wb.BooleanMethods.*;
import static wb.Lck.*;
import static wb.Ent.*;
import static wb.Seg.*;
import static wb.Han.*;
public class Segs {

/* The @var{bsiz} of a segment (given in call to @code{make_seg}) */
/* is a parameter crucial for performance; balancing CPU time traversing */
/* blocks with file-system latency.  @var{bsiz} should be an */
/* integer multiple of the file-system block size. */
/* In the 1990s our nominal @var{bsiz} was 2.kiB; now it should */
/* probably be 4.kiB, 8.kiB, or 16.kiB. */
/*  Segment not in use? */

public static boolean segFree_P(int seg)
{
  if (!(a2b(lckTab)))
    initWb(0x4b, 0x96, 0x800);
  if ((0 > (seg)
      || (seg)>=(numSegs)))
    {
      dprintf(">>>>ERROR<<<< bad segment number "+(seg)+"\n");
      return false;
    }
  else if (!(a2b(seg_Port(seg)))
      && !(a2b(seg_Str(seg)))
      && 0==(seg_Used(seg)))
    return (true);
  else return false;
}


public static boolean checkSeg_P(int seg,String filename)
{
  java.io.RandomAccessFile file = blkio_OpenReadOnlyFile(filename, 0x80);
  byte []tblk = new byte[0x19];
  subbytesRead(tblk, 0, 0x19, file);
  {
    boolean allok_P = 0==(str2long(tblk, 0))
    && 0==(str2long(tblk, 4))
    && 0==(str2long(tblk, 8))
    && 0x77==((tblk[0x17] & 0xFF))
    && 0x62==((tblk[0x18] & 0xFF));
    blkio_FileClose(file);
    return allok_P;
  }
}


public static int errout(String reasonName,int seg,String filename)
{
  dprintf(">>>>ERROR<<<< not a database "+(filename)+" "+(reasonName)+"\n");
  blkio_FileClose(seg_Port(seg));
  doSegBuffers(seg, getMethod("wb.Ents", "recycleEnt", new Class[]{Ent.class}));
  seg_SetPort(seg, null);
  seg_SetStr(seg, null);
  seg_SetUsed(seg, 0);
  return typerr;
}

/* TBD - need to lck seg here. */
/*  Segment will be read-only if MODE is #f. */

/*  */
/* @var{seg} should be a non-negative number.  Opens the database @var{seg} in @var{filename} and */
/* returns @var{seg} if successful, a status code otherwise.  The database @var{seg} */
/* will be read-only if the @var{mode} argument is 0.  It will be read-write if */
/* the @var{mode} argument is 2. */

public static int openSeg(int seg,String filename,int mode)
{
  int bsiz = 3*0x80;
  if (0==(mode))
    mode = 0;
  if (!(segFree_P(seg)))
    {
      dprintf(">>>>ERROR<<<< "+("openSeg")+": segment in use "+(seg)+"\n");
      return argerr;
    }
  else if ((bsiz)>(blkSize))
    {
      dprintf(">>>>ERROR<<<< unsupported "+("bsiz")+" "+(bsiz)+" > "+(blkSize)+"\n");
      return argerr;
    }
  else if (!(checkSeg_P(seg, filename)))
    return  -1;
  else {
    java.io.RandomAccessFile file = a2b(mode)
      ?blkio_OpenModifyFile(filename, bsiz)
      :blkio_OpenReadOnlyFile(filename, bsiz);
Lloop: while (true) {
    if (a2b(mode)
	  ?outputPort_P(file)
	  :inputPort_P(file))
      {
	seg_SetPort(seg, file);
	seg_SetStr(seg, filename);
	seg_SetUsed(seg, 2);
	seg_SetBsiz(seg, bsiz);
	seg_SetFlcLen(seg, a2b(mode)
	    ? -1
	    : -2);
	{
	  Han han = seg_RtHan(seg);
	  byte []tmpStr = new byte[5];
	  if (err_P(btOpen(seg, 0, han, (wcbSap)+(wcbSar))))
	    return errout("bt-open 0", seg, filename);
	  else if (2!=(btGet(han, bsizByts, 4, tmpStr)))
	    return errout("BSIZ", seg, filename);
	  else if ((bsiz)!=(str2short(tmpStr, 0)))
	    {
	      blkio_FileClose(file);
	      bsiz = str2short(tmpStr, 0);
	      if ((bsiz)>(blkSize))
		return errout("BSIZ too big.", seg, filename);
	      else {
		file = a2b(mode)
		  ?blkio_OpenModifyFile(filename, bsiz)
		  :blkio_OpenReadOnlyFile(filename, bsiz);
		continue Lloop;
	      }
	    }
	  else if (4!=(btGet(han, usedByts, 4, tmpStr)))
	    return errout("USED", seg, filename);
	  else {
	    seg_SetUsed(seg, str2long(tmpStr, 0));
	    if (5!=(btGet(han, fldByts, 3, tmpStr)))
	      return errout("FLD", seg, filename);
	    else if (err_P(btOpen(seg, str2long(tmpStr, 1), seg_FlHan(seg), wcbSar)))
	      return errout("FLC", seg, filename);
	    else {
	      if ((han_Typ(seg_FlHan(seg)))!=(frlTyp))
		dprintf("Older type freelist - still supported.\n");
	      han_SetWcb(seg_FlHan(seg), wcbSar);
	      return seg;
	    }
	  }
	}
      }
    else {
      if (a2b(mode)
	    ?inputPort_P(file)
	    :outputPort_P(file))
	blkio_FileClose(file);
      dprintf(">>>>ERROR<<<< could not open file "+(filename)+"\n");
      return ioerr;
    }
    }
  }
}

/*  */
/* Closes database segment @var{seg} and the file containing it.  If @var{hammer_P} is */
/* NULL, then if there are any problems freeing buffers, then the close */
/* is aborted.  A status code is returned. */

public static int closeSeg(int seg,boolean hammer_P)
{
  if ((!(a2b(seg_Str(seg)))
      || 0==(seg_Used(seg))))
    {
      dprintf(">>>>ERROR<<<< "+("closeSeg")+": segment "+(seg)+" already closed\n");
      return argerr;
    }
  else {
    flushFlc(seg, 5);
    /*leave only enough blocks to fit in flc in superblock.*/
    if ((seg_FlcLen(seg))>=0)
      {
	byte []tmpStr = new byte[0x14];
	{
	  int i =  -1+(seg_FlcLen(seg));
	  while (!(0 > (i))) {
	    long2str(tmpStr, 4*(i), seg_Flc(seg)[i]);
	    {
	      i =  -1+(i);
	    }
	  }
	}
	btPut(seg_RtHan(seg), flcByts, 3, tmpStr, 4*(seg_FlcLen(seg)));
	seg_SetFlcLen(seg,  -1);
      }
    {
      int ans = doSegBuffers(seg, getMethod("wb.Ents", "flushBuffer", new Class[]{Ent.class}));
      if ((success_P(ans)
	  || hammer_P))
	{
	  if (!(success_P(ans)))
	    ans = notpres;
	  doSegBuffers(seg, getMethod("wb.Ents", "purgeBuffer", new Class[]{Ent.class}));
	  btClose(seg_RtHan(seg));
	  btClose(seg_FlHan(seg));
	  blkio_FileClose(seg_Port(seg));
	  seg_SetPort(seg, null);
	  seg_SetStr(seg, null);
	  seg_SetUsed(seg, 0);
	}
      return ans;
    }
  }
}


public static byte []dbVersionStr = stringToBytes("wb-2a2");

public static byte []dbAuthorsStr = stringToBytes("A. Jaffer, J. Finger, R. Zito-Wolf");

/*  */
/*  */
/* The integer @var{seg} must specify an unused segment.  The integer @var{bsiz} */
/* specifies the size of B-tree blocks.  @var{bsiz} should be an integer */
/* multiple of the file-system block size.  Nominal value is 4096. */
/*  */
/* @code{makeSeg} makes a new empty database named @var{filename}.  A status code is returned. */
/* Call @code{openSeg} to use the new database. */

public static int makeSeg(int seg,String filename,int bsiz)
{
  if ((!(segFree_P(seg))
      || !(tryLck_P(seg_Lck(seg)))))
    {
      dprintf(">>>>ERROR<<<< "+("makeSeg")+": segment in use "+(seg)+"\n");
      return argerr;
    }
  else if ((bsiz)>(blkSize))
    {
      dprintf(">>>>ERROR<<<< unsupported "+("bsiz")+" "+(bsiz)+" > "+(blkSize)+"\n");
      unlck(seg_Lck(seg));
      return argerr;
    }
  else {
    java.io.RandomAccessFile file = blkio_CreateFile(filename, bsiz);
    if (outputPort_P(file))
      {
	seg_SetPort(seg, file);
	seg_SetBsiz(seg, bsiz);
	seg_SetUsed(seg, 3);
	seg_SetStr(seg, filename);
	initLeafBlk(emptyBlk, 0, dirTyp);
	blk_SetTime(emptyBlk, time(longZero));
	blkio_Write(file, emptyBlk, bsiz, 0);
	initLeafBlk(emptyBlk, 1, dirTyp);
	blk_SetTime(emptyBlk, time(longZero));
	blkio_Write(file, emptyBlk, bsiz, 1);
	initLeafBlk(emptyBlk, 2, frlTyp);
	blk_SetTime(emptyBlk, time(longZero));
	blkio_Write(file, emptyBlk, bsiz, 2);
	blkio_FileClose(file);
	file = blkio_OpenModifyFile(filename, bsiz);
	if (outputPort_P(file))
	  {
	    seg_SetPort(seg, file);
	    unlck(seg_Lck(seg));
	    {
	      Han han = seg_RtHan(seg);
	      byte []tmpStr = new byte[5];
	      btOpen(seg, 0, han, (wcbSap)+(wcbSar));
	      btPut(han, noByts, 0, dbVersionStr, dbVersionStr.length);
	      long2str(tmpStr, 0, seg_Used(seg));
	      btPut(han, usedByts, 4, tmpStr, 4);
	      short2str(tmpStr, 0, seg_Bsiz(seg));
	      btPut(han, bsizByts, 4, tmpStr, 2);
	      tmpStr[0] = (byte) (0xff&4);
	      long2str(tmpStr, 1, 1);
	      btPut(han, rootByts, 4, tmpStr, 5);
	      long2str(tmpStr, 1, 2);
	      btPut(han, fldByts, 3, tmpStr, 5);
	      btPut(han, flcByts, 3, noByts, 0);
	      if ((bsiz)>0x80)
		btPut(han, authorsByts, 7, dbAuthorsStr, dbAuthorsStr.length);
	      return closeSeg(seg, !((true)));
	    }
	  }
	else {
	  dprintf(">>>>ERROR<<<< couldn\'t open fresh file "+(filename)+"\n");
	  unlck(seg_Lck(seg));
	  return ioerr;
	}
      }
    else {
      dprintf(">>>>ERROR<<<< couldn\'t create new file "+(filename)+"\n");
      unlck(seg_Lck(seg));
      return ioerr;
    }
  }
}

/* @noindent The write-control-bits argument (@var{wcb}) to these */
/* functions controls the latency of updates to the file after various */
/* operations.  These bits are defined as follows: */
/*  */
/* @multitable @columnfractions .1 .3 .6 */
/* @item @var{value} */
/* @tab C-name */
/* @tab Meaning */
/* @item 1 */
/* @tab wcb_sap */
/* @tab save block after PUTs */
/* @cindex wcb_sap */
/* @item 2 */
/* @tab wcb_sar */
/* @tab save block after REMOVEs */
/* @cindex wcb_sar */
/* @item 4 */
/* @tab wcb_sac */
/* @cindex wcb_sac */
/* @tab force block save after cached block changes */
/* @item 8 */
/* @tab wcb_fac */
/* @tab flush buffer entirely after cached block changes (not currently */
/* implemented) */
/* @cindex wcb_fac */
/* @end multitable */
/* Opens bt-handle @var{han} to seg number @var{seg}, block number @var{blkNum}, and */
/* returns the type of the block.  If no such block exists or is not a */
/* root block, then a (negative) status code is returned. */

public static int btOpen(int seg,int blkNum,Han han,int wcb)
{
  if ((seg)>=0
      && (seg)<(numSegs)
      && a2b(seg_Str(seg)))
    {
      Ent ent = getEnt(seg, blkNum, accread);
      if (!(a2b(ent)))
	return unkerr;
      else if (!(root_P(ent_Blk(ent))))
	{
	  releaseEnt(ent, accread);
	  dprintf(">>>>ERROR<<<< "+("btOpen")+": not a root "+(seg)+":"+(blkNum)+"\n");
	  return argerr;
	}
      else {
	han_SetSeg(han, seg);
	han_SetNum(han, blkNum);
	han_SetTyp(han, blk_Typ(ent_Blk(ent)));
	han_SetLast(han, blkNum);
	if (blk_Typ_P(ent_Blk(ent), dirTyp))
	  wcb = (wcb)|((wcbSap)+(wcbSar));
	han_SetWcb(han, wcb);
	releaseEnt(ent, accread);
	return han_Typ(han);
      }
    }
  else return argerr;
}

/* Creates a new root block in seg @var{seg} of type @var{typ}, opens bt-handle */
/* @var{han} to it, and returns a status code.  If @var{seg} has insufficient room to */
/* create a new tree, then the @var{noroom} status code is returned. */
/*  */
/* @code{btCreate} can be used to create temporary b-trees.  Temporary trees will be */
/* be reclaimed by check program after system crashes.  In order to */
/* make a tree persistent, add it to a directory (tree). */

public static int btCreate(int seg,int typ,Han han,int wcb)
{
  Ent ent = createNewBlkEnt(seg);
  if (!(a2b(ent)))
    return noroom;
  else {
    int blkNum = ent_Id(ent);
    initLeafBlk(ent_Blk(ent), blkNum, typ);
    ent_SetDty(ent, true);
    ent_SetPus(ent, 0);
    ents_EntWrite(ent);
    han_SetSeg(han, seg);
    han_SetNum(han, blkNum);
    han_SetTyp(han, typ);
    han_SetLast(han, blkNum);
    if ((typ)==(dirTyp))
      wcb = (wcb)|((wcbSap)+(wcbSar));
    han_SetWcb(han, wcb);
    releaseEnt(ent, accwrite);
    return success;
  }
}


/* Closes bt-handle @var{han} and returns @var{SUCCESS}. */
/*  */
/* Currently, @code{btClose} has no effect other than to clear @var{han}. */

public static int btClose(Han han)
{
  han_SetSeg(han, 0);
  han_SetNum(han, 0);
  han_SetTyp(han, 0);
  han_SetLast(han, 0);
  return success;
}

}
