/* "Test.java" WB-tree File Based Associative String Data Base System.
 * Copyright (C) 1991, 1992, 1993, 2000 Free Software Foundation, Inc.
 * Copyright 2007 Clear Methods, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

/* (require 'stdio) */
/* (slib:load-source (in-vicinity (program-vicinity) "wbsys.scm")) */

/* (define diagout stdout) */

/*   call (SCANHAN root-handle [test-prev? #f] [verbose? #t]) */
//package wb;


import static wb.Segs.*;
import static wb.Wbdefs.*;
import static wb.BooleanMethods.*;
import static wb.Wbsys.*;
import static wb.Db.*;
import static wb.Han.*;
import static wb.Seg.*;
import wb.Han;
import java.util.Vector;

public class Test {
    private static boolean stringEqual_P(String a, String b)
    {
        return a.equals(b);
    }

    private static <E> Vector<E> cons(E elem, Vector<E> list)
    {
        list.add(0, elem);
        return list;
    }

    private static  int length(Vector<?> v)
    {
	return v.size();
    }

    public static Vector scanhan(Han han, int ... args)
    {
	int errors = 0;
	int testPrev = 1;//(args.length > 0)? args[0] : 0;
	boolean verbose_P = (args.length > 1)? a2b(args[1]) : false;

	dprintf("FORWARD key scan\n");
	{
	    String key = null;
	    String prior = null;
	    int init = 1;
	    Vector<String> list = new Vector();

	    Lloop: while (true) {
		if (verbose_P)
		    {
			dprintf("NEXT KEY IS ");
			dprintf(key);
			dprintf("\n");
		    }
		if (a2b(testPrev) && !(a2b(init)))
		    {
			String prevKey = bt_Prev(han, key);
			if (a2b(a2b(a2b(prevKey)
				    && a2b(prior)
				    && a2b(!(a2b(stringEqual_P(prevKey, prior)))))
				|| a2b(a2b(prevKey)
				       && a2b(!(a2b(prior))))
				|| a2b(a2b(prior)
				       && a2b(!(a2b(prevKey)))
				       && a2b(!(a2b(stringEqual_P(prior, "")))))))
			    {
				errors = 1+(errors);
				dprintf("NEXT/PREV error: key= ");
				dprintf(prior);
				dprintf("\n");
				dprintf("   next= ");
				dprintf(key);
				dprintf(" prev= ");
				dprintf(prevKey);
				dprintf("\n");
			    }
		    }
		if (a2b(a2b(a2b(key)
			    && a2b(""!=(key)))
			|| a2b(init)))
		    {
			String Tkey = bt_Next(han, key);
			prior = key;
			init = 0;
			list = a2b(key)
			    ?cons(key, list)
			    :list;
			key = Tkey;
			continue Lloop;
		    }
		else {
		    if (a2b(testPrev))
			dprintf("scanhan: " + (errors) + " next/prev errors found.");
		    dprintf("scanhan: " + (length(list)) + " items found.\n");
		    return list;
		}
	    }
	}
    }


    public static int countKeys(Han han)
    {
	{
	    String key = "";
	    int ct = 0;
	    int init = 1;
	    Lloop: while (true) {
		if (a2b(a2b(a2b(key)
			    && a2b(""!=(key)))
			|| a2b(init)))
		    {
			key = bt_Next(han, key);
			ct = (ct)+1;
			init = 0;
			continue Lloop;
		    }
		else return ct;
	    }
	}
    }


    public static Vector scanb(Han han)
    {
	dprintf("REVERSE key scan\n");
	{
	    String key = null;
	    int init = 1;
	    Vector<String> list = new Vector();

	    Lloop: while (true) {
		dprintf("PREV KEY IS ");
		dprintf(key);
		dprintf("\n");
		if (a2b(a2b(a2b(key)
			    && a2b(""!=(key)))
			|| a2b(init)))
		    {
			String Tkey = bt_Prev(han, key);
			init = 0;
			list = a2b(key)
			    ?cons(key, list)
			    :list;
			key = Tkey;
			continue Lloop;
		    }
		else {
		    dprintf("%s: %d items found.\n"+("scanb")+(length(list)));
		    return list;
		}
	    }
	}
    }


    public static int dbSize(Han han)
    {
	return (seg_Used(han_Seg(han)))+1;
    }


    public static Han currentBt = null;

    public static int currentSeg = 0;

    public static int addKeyNum = 0;

    public static String addStr = "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyz";

    /*  ADD! count [first-key add-key-num] [key-increment 1]  [value-string "abc...xyz"x2] */


    public static void add(int n, int ... args)
    {
	int curKeyNum =(args.length > 0)? args[0] : addKeyNum;
	int incr = (args.length > 1)?  args[1] : 1;
	String addstr = (args.length > 2)?  args[2] + "" : addStr;

        int i = 1;
	while (!(i > n)) {
	    String keyStr = curKeyNum + "";
	    curKeyNum = (incr)+(curKeyNum);
	    bt_Put(currentBt, keyStr, addstr);
	    i = (i)+1;
	}
	addKeyNum = max(addKeyNum, curKeyNum);
    }

    /*  REMOVE! count first-key [key-increment 1] [unused-arg] */


    public static void remove(int n, int curKey, int ... args)
    {
	int incr = a2b(args.length > 0)? args[0]:1;
// 	int startTrace = (args.length > 1)? args[1]:0;

        int i = 1;
	while (!(i > n)) {
	    String keyStr = curKey + "";
	    curKey = (incr)+(curKey);
	    bt_Rem(currentBt, keyStr);
	    i = (i)+1;
	}
	return;
    }

    /*  old test code, still useful! */


    public static void test0(Han bHan)
    {
	bt_Put(bHan, "foo", "bar");
	bt_Put(bHan, "foz", "oof");
	bt_Put(bHan, "fonz", "zonf");
	bt_Put(bHan, "foo", "raboof");
	bt_Put(bHan, "food", "thought");
	scanhan(bHan);
	bt_Rem(bHan, "foo");
	bt_Rem(bHan, "fonz");
	bt_Put(bHan, "foo", "foo=bar");
	bt_Put(bHan, "foz", "foz=oof");
	bt_Put(bHan, "fonz", "fonz=zonf");
	bt_Put(bHan, "foo", "foo=raboof");
	bt_Put(bHan, "food", "food=for thought");
	bt_Put(bHan, "very very very long key field", "very very very longer value field");
	scanhan(bHan);
    }

    /*  THIS IS TEST CASE #1. */
    /*  It is used to test the addition of data elements with similar names */
    /*  to see if the repeat count routine is working correctly */

    public static void test1()
    {
	Han aHan = createBt(currentSeg, indTyp, 0);
	bt_Put(aHan, "cat", "6");
	bt_Put(aHan, "caddy", "4");
	bt_Put(aHan, "catalytic", "7");
	bt_Put(aHan, "cadalack", "3");
	bt_Put(aHan, "catastrophy", "10");
	bt_Put(aHan, "catastrophic", "9");
	bt_Put(aHan, "cad", "2");
	bt_Put(aHan, "cadjole", "5");
	bt_Put(aHan, "cataract", "8");
	bt_Put(aHan, "cacky", "1");
	bt_Put(aHan, "d", "12");
	scanhan(aHan);
	dprintf("   TEST 1: nexting on block 'a' index 'cadz'\n");
	{
	    String ansStr = bt_Next(aHan, "cadz");
	    dprintf(ansStr);
	}
	dprintf("   TEST 1: removing 'd' from 'a'\n");
	bt_Rem(aHan, "d");
	scanhan(aHan);
	dprintf("   TEST 1: removing 'cad' from 'a'\n");
	bt_Rem(aHan, "cad");
	scanhan(aHan);
	closeBt(aHan);
    }

    /*  THIS IS TEST #2 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block completely then adds a new data element and value */
    /*  to the block to make it split */

    public static void test2()
    {
	Han bHan = createBt(currentSeg, indTyp, 0);
	bt_Put(bHan, "foo", "bar");
	bt_Put(bHan, "foz", "oof");
	bt_Put(bHan, "fonz", "zonf");
	bt_Put(bHan, "foo", "raboof");
	bt_Put(bHan, "food", "thought");
	scanhan(bHan);
	bt_Rem(bHan, "foo");
	bt_Rem(bHan, "fonz");
	scanhan(bHan);
	dprintf("   TEST 2: creating a full block named 'b'\n");
	bt_Put(bHan, "foo", "1234567");
	bt_Put(bHan, "foz", "12345678901");
	bt_Put(bHan, "fonz", "123456789");
	bt_Put(bHan, "foo", "1234567890");
	bt_Put(bHan, "food", "1234567890123456");
	bt_Put(bHan, "test", "123456789012345678901234");
	dprintf("   TEST 2: doing a PUT to fill the block\n");
	bt_Put(bHan, "a", "12345678");
	scanhan(bHan);
	closeBt(bHan);
    }

    /*  THIS IS TEST #3 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block , then it tries to make the block split by making */
    /*  the first data element 1 char bigger by replacement */

    public static void test3()
    {
	Han cHan = createBt(currentSeg, indTyp, 0);
	bt_Put(cHan, "foo", "1234567");
	bt_Put(cHan, "foz", "12345678901");
	bt_Put(cHan, "fonz", "123456789");
	bt_Put(cHan, "foo", "1234567890");
	bt_Put(cHan, "food", "1234567890123456");
	bt_Put(cHan, "test", "123456789012345678901234");
	bt_Put(cHan, "a", "12345678");
	scanhan(cHan);
	dprintf("   TEST 3: split block by increasing first element by one char with PUT\n");
	bt_Put(cHan, "a", "123456789");
	scanhan(cHan);
	closeBt(cHan);
    }

    /*  THIS IS TEST #4 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block , then it tries to make the block split by making */
    /*  a middle data element 1 char bigger by replacement */

    public static void test4()
    {
	Han dHan = createBt(currentSeg, indTyp, 0);
	bt_Put(dHan, "foo", "1234567");
	bt_Put(dHan, "foz", "12345678901");
	bt_Put(dHan, "fonz", "123456789");
	bt_Put(dHan, "foo", "1234567890");
	bt_Put(dHan, "food", "1234567890123456");
	bt_Put(dHan, "test", "123456789012345678901234");
	bt_Put(dHan, "a", "12345678");
	scanhan(dHan);
	dprintf("   TEST 4: split block by incresing value of a middle element by one with PUT\n");
	bt_Put(dHan, "foz", "123456789012");
	scanhan(dHan);
	closeBt(dHan);
    }

    /*  THIS IS TEST #5 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block , then it tries to make the block split by making */
    /*  the last data element 1 char bigger by replacement */

    public static void test5()
    {
	Han eHan = createBt(currentSeg, indTyp, 0);
	bt_Put(eHan, "foo", "1234567");
	bt_Put(eHan, "foz", "12345678901");
	bt_Put(eHan, "fonz", "123456789");
	bt_Put(eHan, "foo", "1234567890");
	bt_Put(eHan, "food", "1234567890123456");
	bt_Put(eHan, "test", "123456789012345678901234");
	bt_Put(eHan, "a", "12345678");
	scanhan(eHan);
	dprintf("TEST 5: splitblock by increasing value of the \n"
		+ "last data element by one with PUT\n");
	bt_Put(eHan, "test", "12345678901234567890123456");
	scanhan(eHan);
	closeBt(eHan);
    }

    /*  THIS IS TEST #6 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block , then it deletes the first element. It then */
    /*  tries to make the block split by reentering the first data element */
    /*  with 1 more char than the original */

    public static void test6()
    {
	Han fHan = createBt(currentSeg, indTyp, 0);
	bt_Put(fHan, "foo", "1234567");
	bt_Put(fHan, "foz", "12345678901");
	bt_Put(fHan, "fonz", "123456789");
	bt_Put(fHan, "foo", "1234567890");
	bt_Put(fHan, "food", "1234567890123456");
	bt_Put(fHan, "test", "123456789012345678901234");
	bt_Put(fHan, "a", "12345678");
	scanhan(fHan);
	dprintf("   TEST 6: split block by deleting the first data element with REM\n");
	dprintf("   TEST 6: then reentering the first data element with 1 more char\n");
	bt_Rem(fHan, "a");
	dprintf("   TEST 6: doing a PUT\n");
	bt_Put(fHan, "a", "123456789");
	scanhan(fHan);
	closeBt(fHan);
    }

    /*  THIS IS TEST #7 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block , then it deletes a middle element. It then */
    /*  tries to make the block split by reentering a middle element with 1 */
    /*  more char than the original */

    public static void test7()
    {
	Han gHan = createBt(currentSeg, indTyp, 0);
	bt_Put(gHan, "foo", "1234567");
	bt_Put(gHan, "foz", "12345678901");
	bt_Put(gHan, "fonz", "123456789");
	bt_Put(gHan, "foo", "1234567890");
	bt_Put(gHan, "food", "1234567890123456");
	bt_Put(gHan, "test", "123456789012345678901234");
	bt_Put(gHan, "a", "12345678");
	scanhan(gHan);
	dprintf("   TEST 7: split block by deleting a middle data element\n");
	dprintf("   TEST 7: then reentering a middle data element with 1 more char\n");
	dprintf("   TEST 7: than the original value.\n");
	bt_Rem(gHan, "foo");
	dprintf("   TEST 7: doing a PUT\n");
	bt_Put(gHan, "foo", "12345678901");
	scanhan(gHan);
	closeBt(gHan);
    }

    /*  THIS IS TEST #8 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block , then it deletes the last element. It then */
    /*  tries to make the block split by reentering the last element with 1 */
    /*  more char than the original */

    public static void test8()
    {
	Han hHan = createBt(currentSeg, indTyp, 0);
	bt_Put(hHan, "foo", "1234567");
	bt_Put(hHan, "foz", "12345678901");
	bt_Put(hHan, "fonz", "123456789");
	bt_Put(hHan, "foo", "1234567890");
	bt_Put(hHan, "food", "1234567890123456");
	bt_Put(hHan, "test", "123456789012345678901234");
	bt_Put(hHan, "a", "12345678");
	scanhan(hHan);
	dprintf("   TEST 8: split block by deleting the last data element\n");
	dprintf("   TEST 8: then reentering the last data element with 1 more char\n");
	dprintf("   TEST 8: than the original value.  Block name is 'h'\n");
	bt_Rem(hHan, "test");
	dprintf("   TEST 8: doing a PUT\n");
	bt_Put(hHan, "test", "12345678901234567890123456");
	scanhan(hHan);
	closeBt(hHan);
    }

    /*  THIS IS TEST #9 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block completely then adds a new data element and value */
    /*  to the block to make it split */

    public static void test9()
    {
	Han iHan = createBt(currentSeg, indTyp, 0);
	bt_Put(iHan, "foz", "12345678901");
	bt_Put(iHan, "fonz", "123456789");
	bt_Put(iHan, "foo", "1234567890");
	bt_Put(iHan, "food", "1234567890123456");
	bt_Put(iHan, "test", "123456789012345678901234");
	bt_Put(iHan, "a", "12345678");
	dprintf("   TEST 9: The block 'i' is full. Now we are adding a new index to the begining\n");
	dprintf("   TEST 9: of the block with a value that should make the block split\n");
	bt_Put(iHan, "1", "123456789");
	scanhan(iHan);
	closeBt(iHan);
    }

    /*  THIS IS TEST #10 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block completely then adds a new data element and value */
    /*  to the block to make it split */

    public static void test10()
    {
	Han jHan = createBt(currentSeg, indTyp, 0);
	bt_Put(jHan, "foz", "12345678901");
	bt_Put(jHan, "fonz", "123456789");
	bt_Put(jHan, "foo", "1234567890");
	bt_Put(jHan, "food", "1234567890123456");
	bt_Put(jHan, "test", "123456789012345678901234");
	bt_Put(jHan, "a", "12345678");
	dprintf("   TEST 10: The block 'j' is full. Now we are adding a new index to the middle\n");
	dprintf("   TEST 10: of the block with a value that should make the block split\n");
	bt_Put(jHan, "fooa", "123456789");
	scanhan(jHan);
	closeBt(jHan);
    }

    /*  THIS IS TEST #11 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block completely then adds a new data element and value */
    /*  to the block to make it split */

    public static void test11()
    {
	Han kHan = createBt(currentSeg, indTyp, 0);
	bt_Put(kHan, "foz", "12345678901");
	bt_Put(kHan, "fonz", "123456789");
	bt_Put(kHan, "foo", "1234567890");
	bt_Put(kHan, "food", "1234567890123456");
	bt_Put(kHan, "test", "123456789012345678901234");
	bt_Put(kHan, "a", "12345678");
	dprintf("   TEST 11: The block 'k' is full. Now we are adding a new index to the end\n");
	dprintf("   TEST 11: of the block with a value that should make the block split\n");
	bt_Put(kHan, "zzz", "123456789");
	scanhan(kHan);
	closeBt(kHan);
    }

    /*  THIS IS TEST #12 FOR BLOCK SPLITTING STUFF */
    /*  this test fills a block completely then adds a new data element and value */
    /*  to the block to make it split */

    public static void test12()
    {
	Han kHan = createBt(currentSeg, indTyp, 0);
	bt_Put(kHan, "132", "12345678901");
	bt_Put(kHan, "1233", "1234567890");
	bt_Put(kHan, "26", "1234567890123456");
	bt_Put(kHan, "275", "123456789012345678901234");
	bt_Put(kHan, "84", "12345678");
	dprintf("TEST 12: The block 'k' is full. Now we are \n" +
		"adding a new index pastp after b-pos\n");
	dprintf("   TEST 12: of the block with a value that should make the block split\n");
	bt_Put(kHan, "82", "123456789");
	scanhan(kHan);
	closeBt(kHan);
    }


    public static void test13()
    {
	Han kHan = createBt(currentSeg, indTyp, 0);
	bt_Put(kHan, "0132", "12345678901");
	bt_Put(kHan, "01233", "1234567890");
	bt_Put(kHan, "026", "1234567890123456");
	bt_Put(kHan, "0275", "123456789012345678901234");
	bt_Put(kHan, "084", "12345678");
	dprintf("TEST 13: The block 'k' is full. Now we are adding" +
		" a new index pastp after b-pos\n");
	dprintf("   TEST 13: of the block with a value that should make the block split\n");
	bt_Put(kHan, "082", "123456789");
	scanhan(kHan);
	closeBt(kHan);
    }


    //     public static void test14()
    //     {
    //	Han kHan = createBt(currentSeg, indTyp, 0);
    //	bt_Put(kHan, "0132", "12345678901");
    //	bt_Put(kHan, "01233", "1234567890");
    //	bt_Put(kHan, "026", "1234567890123456");
    //	bt_Put(kHan, "04", "123456789012345678901234");
    //	bt_Put(kHan, "041", "1234567890");
    //	dprintf("\nTEST 14: The  block 'k' is full. Now we are adding\n" +
    //             "a new index pastp after b-pos\n");
    //	dprintf("   TEST 14: of the block with a value that should make the block split\n");
    //	tscanBlk(id(kHan));
    //	bt_Put(kHan, "040", "123456789");
    //	tscanBlk(id(kHan));
    //	closeBt(kHan);
    //     }


    public static void test128()
    {
	test1();
	test2();
	test3();
	test4();
	test5();
	test6();
	test7();
	test8();
	test9();
	test10();
	test11();
	test12();
	test13();
    }


    public static void main(String[] args)
    {
	dprintf("make-seg\n");
	makeSeg(9, "z", 0x80);
	dprintf("open-seg\n");
	currentSeg = openSeg(9, "z", 2);
	dprintf("create-bt\n");
	currentBt = createBt(currentSeg, indTyp, 0);
	dprintf("test0!\n");
	test0(currentBt);
	dprintf("test!\n");
	test128();
	dprintf("add! 10\n");
	add(0xa);
	closeSeg(currentSeg, false);
	currentSeg = openSeg(9, "z", 2);
	scanhan(currentBt);
	closeSeg(currentSeg, false);
    }


    //     public static int dbview(String filename)
    //     {
    //	currentSeg = openSeg(9, filename, 0);
    //	{
    //	    int blknum = read();
    //	    while (a2b(!(a2b(eofObject_P(blknum))))) {
    //		tscanBlk(blknum);
    //		{
    //		    blknum = read();
    //		}
    //	    }
    //	    return closeSeg(currentSeg, null);
    //	}
    //     }


    //     public static int rmain(int args)
    //     {
    //	dprintf("make-seg\n");
    //	makeSeg(9, "z", 0x80);
    //	dprintf("open-seg\n");
    //	currentSeg = openSeg(9, "z", 2);
    //	dprintf("create-bt\n");
    //	currentBt = createDb(currentSeg, indTyp, "test");
    //	dprintf("tscan \n");
    //	tscan();
    //	dprintf("test0!\n");
    //	test0(currentBt);
    //	tscan();
    //	if (a2b((length(args))>0))
    //	    {
    //		dprintf("TESTING for INSERT BUG\n");
    //		bt_Put(currentBt, "xxx2",
    //		    "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz");
    //		tscan();
    //		bt_Put(currentBt, "xxx3",
    //		    "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz");
    //		tscan();
    //		bt_Put(currentBt, "xxx1",
    //		    "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz");
    //		tscan();
    //		bt_Put(currentBt, "xxx0",
    //		    "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz");
    //		tscan();
    //	    }
    //	if (a2b(false))
    //	    {
    //		dprintf("add! 10\n");
    //		add(0xa);
    //	    }
    //	closeSeg(currentSeg, false);
    //	currentSeg = openSeg(9, "z", 2);
    //	currentBt = openDb(currentSeg, "test");
    //	scanhan(currentBt);
    //	closeSeg(currentSeg, false);
    //	currentSeg = 0;
    //	return 0;
    //     }

    //     /*  Aubrey -- I expect this routine to leave the file OPEN --roy. */
    //     //TODO: to see this again
    //     public static void qmain()
    //     {
    //	dprintf("open-seg\n");
    //	currentSeg = openSeg(9, "z", 2);

    //	Han han = openDb(currentSeg, "test");
    //	if (!(a2b(han))) {
    //	   dprintf("`test' created\n");
    //            createDb(9, indTyp, "test");
    //            currentBt = null;
    //        } else
    //            currentBt = han;

    //	if (a2b(currentBt))
    //	    scanhan(currentBt);
    //	else
    //         dprintf("db `test' not found\n");
    //     }

    //     /*  The rest of this will run only in scheme. */

    //     /*  call (TSCAN-FULL root-seg root-blk */
    //     /*               [#of-sublevels 1] [#of-blks-to-chain-fwd 99] [mode 2] */
    //     /*   (plus internal args:) */
    //     /*               [indent-string ''] [split-key #f] [skey-len 0] */
    //     /*               [parent -1] [expected-level -1] */
    //     /*  MODE: 0=print errors only; 1=errors+warnings; 2=1+block contents */
    //     /*  */
    //     /*  call (TSTATS) for just error check and tree stats */
    //     /*  */
    //     /*  BUG: subblocks missing ptr from parent wont show up! (what a pain) */
    //     /*  BUG: especially on the right fringe of the tree! */
    //     /*  BUG: doesnt yet recurse through directories */
    //     /*  */
    //     /*  Instrumented to test five things: */
    //     /*   1. that keys are in order at each level; */
    //     /*   2. that keys are carried over correctly between levels; */
    //     /*   3. to flag blocks missing their parent pointer (not an error, really) */
    //     /*   4+5. check consistency of ROOT PTR and TREE TYPE */
    //     /*  */
    //     /*  Instrumented to collect statistics: */
    //     /*   A. Blocks used (per level; overall) */
    //     /*   B. Branching factor (per level; overall) */
    //     /*   C. Block utilization (per level; overall) */
    //     /*  */

    //     public static int tscanFull(int seg, int blkNum, int args)
    //     {
    //	{
    //	    int nlevels = a2b((length(args))>0)
    //		?car(args)
    //		:1;
    //	    int nblks = a2b((length(args))>1)
    //		?cadr(args)
    //		:0x63;
    //	    int mode = a2b((length(args))>2)
    //		?caddr(args)
    //		:2;
    //	    rootBlk = 0;
    //	    rootType = 0;
    //	    tscanErrs = 0;
    //	    tscanWarns = 0;
    //	    {
    //		int i = 0;
    //		while (a2b(!(a2b((i)>=0x14)))) {
    //		    tscanKeycts[i] = 0;
    //		    tscanBlkcts[i] = 0;
    //		    tscanBytects[i] = 0;
    //		    tscanLengths[i] =  -1;
    //		    tscanPrevs[i] =  -1;
    //		    tscanExpnexts[i] =  -1;
    //		    {
    //			i = (i)+1;
    //		    }
    //		}
    //	    }
    //	    tscanInternal(seg, blkNum, nlevels, nblks, mode, "", "", 0,  -1,  -1);
    //	    dprintf("\n%d ERRORS, %d WARNINGS.\n\n"+(tscanErrs)+(tscanWarns));
    //	    dprintf("TREE SHAPE STATISTICS\n");
    //	    dprintf("LEVEL  BLOCKS    KEYS BRANCHING  BYTES  %%USED \n");
    //	    dprintf("                 /BLK FACTOR     /BLK   /BLK  \n");
    //	    {
    //		int tblks = 0;
    //		int tkeys = 0;
    //		int tbytes = 0;
    //		int bpb = (bsiz(seg))-(blkDataStart);
    //		{
    //		    int i = 0;
    //		    while (a2b(!(a2b((i)>=0x14)))) {
    //			{
    //			    int blks = tscanBlkcts[i];
    //			    int keys = tscanKeycts[i];
    //			    int bytes = tscanBytects[i];
    //			    tblks = (tblks)+(blks);
    //			    tkeys = (tkeys)+(keys);
    //			    tbytes = (bytes)+(tbytes);
    //			    if (a2b((blks)>0))
    //				dprintf("%5d %7d %7d %8d%% %6d %6d%%\n"+(i)+(blks)+(keys)+(((keys)*0x64)/(blks))+((((bytes)/(blks))*(bpb))/0x64)+((bytes)/(blks)));
    //			}
    //			{
    //			    i = (i)+1;
    //			}
    //		    }
    //		}
    //		dprintf("TOTAL %7d %7d %8d%% %6d %6d%%\n"+(tblks)+(tkeys)+(((tkeys)*0x64)/(tblks))+((((tbytes)/(tblks))*(bpb))/0x64)+((tbytes)/(tblks)));
    //		return dprintf("\n BLKS-USED=%d BLK-SIZE=%d FILE-SIZE=%d\n"+((used(seg))+1)+(bsiz(seg))+(((used(seg))+1)*(bsiz(seg))));
    //	    }
    //	}
    //     }


    //     public static int tscan()
    //     {
    //	checkAccess();
    //	return tscanBlk(id(currentBt));
    //     }


    //     public static int tstats()
    //     {
    //	checkAccess();
    //	return tscanFull(currentSeg, id(currentBt), 0x63, 0x63, 1);
    //     }


    //     public static byte []tscanBlk(int blkNum)
    //     {
    //	checkAccess();
    //	return tscanFull(currentSeg, blkNum);
    //     }


    //     public static byte []tmpStr2 = new byte[0x100];

    //     public static byte []tmpStr3 = new byte[0x100];

    //     public static int tscanErrs = 0;

    //     public static int tscanWarns = 0;

    //     public static byte []rootBlk = null;

    //     public static int rootType = 0;

    //     public static int [] tscanKeys = new int[0x14];

    //     public static int [] tscanLengths = new int[0x14];

    //     public static int [] tscanPrevs = new int[0x14];

    //     public static int [] tscanExpnexts = new int[0x14];

    //     public static int []tscanKeycts = new int[0x14];

    //     public static int []tscanBlkcts = new int[0x14];

    //     public static int [] tscanBytects = new int[0x14];



    //     public static int tscanInternal(int seg, int blkNum, int nlevels, int nblks, int mode, Ent indent, int skey, int skeyLen, Ent parent, int explevel)
    //     {
    //	LtscanInternal:while (true) {
    //	    {
    //		Ent ent = getEnt(seg, blkNum, accread);
    //		byte []blk = a2b(ent)
    //		    ?blk(ent)
    //		    :null;
    //		int blklev = a2b(ent)
    //		    ?(level(blk))-(leaf)
    //		    :
    //		    int tlen = reconThisKeyDebug(blk, splitKeyPos(blk), tmpStr2, 0, 0x100);
    //		int lidx = (level(blk))-(leaf);
    //		byte []lastKeyStr = tscanKeys[lidx];
    //		int lastLen = tscanLengths[lidx];
    //		if (a2b(ent))
    //		    {
    //			if (a2b((parent)== -1))
    //			    {
    //				rootBlk = topId(blk);
    //				rootType = typ(blk);
    //			    }
    //			tscanBlkcts[blklev] = (tscanBlkcts[blklev])+1;
    //			if (a2b((mode)>1))
    //			    dprintf("\
	// %s scanning blk \
	// %d:%ld top= %d n\
	// ext= %d len= %d \
	// room= %d level= \
	// %d type= %c end=\
	//  %d\n"+(indent)+(seg(ent))+(id(blk))+(topId(blk))+(nxtId(blk))+((end(blk))-(blkDataStart))+((bsiz(currentSeg))-(end(blk)))
	//				    +(level(blk))+((typ(blk)))+(end(blk)));
	//			if (a2b((rootBlk)!=(topId(blk))))
	//			    {
	//				tscanErrs = (tscanErrs)+1;
	//				dprintf(">>>>ERROR<<<< Block "+(seg(ent))+" in tree "+(id(ent))+" thinks its in tree "+(rootBlk)+"");
	//			    }
	//			if (a2b((rootType)!=(typ(blk))))
	//			    {
	//				tscanErrs = (tscanErrs)+1;
	//				dprintf(">>>>ERROR<<<< Block "+(seg(ent))+" in tree type "+(id(ent))+" thinks its of type "+(rootType)+"");
	//			    }
	//			if (a2b(a2b((parent)> -1)
	//				&& a2b(!(a2b(strEqual_P(tmpStr2, 0, tlen, skey, 0, skeyLen))))))
	//			    {
	//				tscanErrs = (tscanErrs)+1;
	//				{
	//				    boolean error_P = strGtr_P(tmpStr2, 0, tlen, skey, 0, skeyLen);
	//				    if (a2b(error_P))
	//					{
	//					    dprintf("ERROR split key mismatch:\n");
	//					    tscanErrs = (tscanErrs)+1;
	//					}
	//				    else {
	//					tscanWarns = (tscanWarns)+1;
	//					if (a2b((mode)>0))
	//					    dprintf("WARNING: split key mismatch:\n");
	//				    }
	//				    if (a2b(a2b(error_P)
	//					    || a2b((mode)>0)))
	//					{
	//					    dprintf("  --Block %d:%ld has key %.*s (%d)\n"+(seg)+(blkNum)+(tlen)+(tmpStr2)+(tlen));
	//					    dprintf("  --while parent %d:%ld thinks key is %.*s (%d)\n"+(seg)+(parent)+(skeyLen)+(skey)+(skeyLen));
	//					}
	//				}
	//			    }
	//			{
	//			    int expnext = tscanExpnexts[lidx];
	//			    int prev = tscanPrevs[lidx];
	//			    if (a2b(a2b((expnext)> -1)
	//				    && a2b((expnext)!=(blkNum))))
	//				{
	//				    warnsErrs = (tscanWarns)+1;
	//				    if (a2b((mode)>0))
	//					{
	//					    dprintf("WARNING: parent ptr missing for blk "+(seg)+"");
	//					    dprintf(" --current blk=%d, parent=%d, last=%d, exp=%d.\n"+(blkNum)+(parent)+(prev)+(expnext));
	//					}
	//				}
	//			}
	//			if (a2b(a2b((explevel)> -1)
	//				&& a2b((explevel)!=(level(blk)))))
	//			    {
	//				tscanErrs = (tscanErrs)+1;
	//				dprintf(">>>>ERROR<<<< blk "+(seg)+" at level "+(blkNum)+" is child of blk "+(level(blk))+" at level "+(seg)
	//					+"");
	//			    }
	//			tscanExpnexts[lidx] = nxtId(blk);
	//			tscanPrevs[lidx] = blkNum;
	//			{
	//			    int firstKey = 1;
	//			    int bPos = blkDataStart;
	//			    int pos = 0;
	//			    int count = 0;
	//			    byte []keyStr = tmpStr2;
	//			    int kLen = 0;
	//			    while (a2b(!(a2b((bPos)>=(end(blk)))))) {
	//				pos = bPos;
	//				kLen = reconThisKeyDebug(blk, pos, keyStr, 0, 0x100);
	//				bPos = nextField(blk, 1+(bPos));
	//				if (a2b(a2b((lastLen)> -1)
	//					&& a2b(a2b(firstKey)
	//					       ?strGtr_P(lastKeyStr, 0, lastLen, keyStr, 0, kLen)
	//					       :!(a2b(strGtr_P(keyStr, 0, kLen, lastKeyStr, 0, lastLen))))))
	//				    {
	//					tscanErrs = (tscanErrs)+1;
	//					dprintf("KEY ORDER ERROR at blk %d:%ld pos %d\n"+(seg)+(blkNum)+(pos));
	//					dprintf(" --key=%.*s last key=%.*s.\n"+(kLen)+(keyStr)+(lastLen)+(lastKeyStr));
	//				    }
	//				substringMove(keyStr, 0, kLen, lastKeyStr, 0);
	//				tscanLengths[lidx] = kLen;
	//				lastLen = kLen;
	//				if (a2b((bPos)<(end(blk))))
	//				    {
	//					if (a2b(leaf_P(blk)))
	//					    {
	//						if (a2b((mode)>1))
	//						    dprintf("%s at %d key= %s value= %s\n"+(indent)+(pos)+(keyStr)+(&blk[1+(bPos)]));
	//					    }
	//					else {
	//					    if (a2b((mode)>1))
	//						dprintf("%s at %d key= %s ptr= %d\n"+(indent)+(pos)+(keyStr)+(str2long(blk, 1+(bPos))));
	//					    if (a2b((nlevels)>0))
	//						{
	//						    int TblkNum = str2long(blk, 1+(bPos));
	//						    nlevels = (nlevels)-1;
	//						    nblks = 0;
	//						    indent = stringAppend(indent, "   ");
	//						    skey = keyStr;
	//						    skeyLen = kLen;
	//						    parent = blkNum;
	//						    explevel = (level(blk))-1;
	//						    blkNum = TblkNum;
	//						    continue LtscanInternal;
	//						}
	//					}
	//					bPos = nextField(blk, bPos);
	//				    }
	//				else {
	//				    tscanKeycts[blklev] = (tscanKeycts[blklev])+(count);
	//				    tscanBytects[blklev] = (tscanBytects[blklev])+((0x64*((end(blk))-(blkDataStart)))/((bsiz(seg))-(blkDataStart)));
	//				    if (a2b((mode)>1))
	//					dprintf("%s at %d split= %s\n"+(indent)+(pos)+(keyStr));
	//				}
	//				{
	//				    firstKey = 0;
	//				    count = (count)+1;
	//				}
	//			    }
	//			}
	//			{
	//			    int nxt = nxtId(blk);
	//			    releaseEnt(ent, accread);
	//			    if (a2b(a2b(nxt)
	//				    && a2b((nblks)>0)))
	//				{
	//				    dprintf("\n");
	//				    {
	//					blkNum = nxt;
	//					nblks = (nblks)-1;
	//					skey = "";
	//					skeyLen = 0;
	//					continue LtscanInternal;
	//				    }
	//				}
	//			    else return 0;
	//			}
	//		    }
	//		else {
	//		    tscanErrs = (tscanErrs)+1;
	//		    return dprintf("%s ERROR: can't access blk %d:%ld. \n"+(indent)+(seg)+(blkNum));
	//		}
	//	    }
	//	}
	//     }


	//     public static boolean strEqual_P(byte []aStr, int aPos, int aLen, byte []bStr, int bPos, int bLen)
	//     {
	//	return a2b((aLen)==(bLen))
	//	    && a2b({
	//		    int i = 0;
	//		    int ap = aPos;
	//		    int bp = bPos;
	//		    Lloop: while (true) {
	//			a2b((i)>=(aLen))
	//			    ?(true)
	//			    :a2b((aStr[ap])!=(bStr[bp]))
	//			    ?false
	//			    :loop((i)+1, (ap)+1, (bp)+1)
	//			    }
	//		});
	//     }


	//     /*  temp hack in case rean RECON is traced... */


	//     public static int reconThisKeyDebug(byte []blk, int pos, byte []keyStr, int kPos, int kLen)
	//     {
	//	{
	//	    int bPos = blkDataStart;
	//	    int kSize = 0;
	//	    while (a2b(!(a2b((bPos)>(pos))))) {
	//		substringMove(blk, (bPos)+2, (bPos)+2+(fieldLen(blk, 1+(bPos))), keyStr, (kPos)+(fieldLen(blk, bPos)));
	//		kSize = (fieldLen(blk, bPos))+(fieldLen(blk, 1+(bPos)));
	//		if (a2b((kSize)>=(kLen)))
	//		    ;
	//		bPos = nextField(blk, 1+(bPos));
	//		if (a2b((bPos)<(end(blk))))
	//		    bPos = nextField(blk, bPos);
	//		{
	//		}
	//	    }
	//	    &keyStr[kPos];
	//	    return kSize;
	//	}
	//     }

	//     /*  (close-seg 9 #f) (load "all") (load "main") (rmain)(qmain)(cstats)(add! 50) (add! 300 "a b c d e f g h i j")  (Cstats) */


	//     public static int prof()
	//     {
	//	makeSeg(9, "z", 0x800);
	//	currentSeg = openSeg(9, "z", 2);
	//	currentBt = createDb(9, indTyp, "test");
	//	clearStats();
	//	start = getInternalRunTime();
	//	add(0x64, 0, 1);
	//	add(0x64, 0x3de,  -10);
	//	add(0x64, 0x3e8, 0x64);
	//	display("Adds took ");
	//	display((0x3e8*((getInternalRunTime())-(start)))/(internalTimeUnitsPerSecond));
	//	display(" Msec");
	//	dprintf("\n");
	//	cstats();
	//	start = getInternalRunTime();
	//	scanhan(currentBt);
	//	display("forward scan took ");
	//	display((0x3e8*((getInternalRunTime())-(start)))/(internalTimeUnitsPerSecond));
	//	display(" Msec");
	//	dprintf("\n");
	//	cstats();
	//	start = getInternalRunTime();
	//	scanb(currentBt);
	//	display("backward scan took ");
	//	display((0x3e8*((getInternalRunTime())-(start)))/(internalTimeUnitsPerSecond));
	//	display(" Msec");
	//	dprintf("\n");
	//	cstats();
	//	start = getInternalRunTime();
	//	remove(0x64, 0, 1);
	//	remove(0x64, 0x3de,  -10);
	//	remove(0x64, 0x3e8, 0x64);
	//	display("Removes took ");
	//	display((0x3e8*((getInternalRunTime())-(start)))/(internalTimeUnitsPerSecond));
	//	display(" Msec");
	//	dprintf("\n");
	//	cstats();
	//	closeSeg(currentSeg, null);
	//	return 0;
	//	Lstart:while (true)
	//	    return 0;
	//     }


	//     public static void radd(int span)
	//     {
	//	require("random");
	//	dprintf("adding %d records with random keys\n"+(span));
	//	{
	//	    int i = addKeyNum;
	//	    int r = random(0xa*(span));
	//	    while (a2b(!(a2b((i)>=((addKeyNum)+(span)))))) {
	//		bt_Put(currentBt, numberToString(r), stringAppend(numberToString(i), "number written in random test"));
	//		{
	//		    i = 1+(i);
	//		    r = random(0xa*(span));
	//		}
	//	    }
	//	}
	//	addKeyNum = (addKeyNum)+(span);
	//     }


	//     public static void radd(int span, int key)
	//     {
	//	require("random");
	//	dprintf("adding %d records with random keys\n"+(span));
	//	{
	//	    int i = addKeyNum;
	//	    int r = random(0xa*(span));
	//	    while (a2b(!(a2b((i)>=((addKeyNum)+(span)))))) {
	//		bt_Put(currentBt, numberToString(r), key);
	//		{
	//		    i = 1+(i);
	//		    r = random(0xa*(span));
	//		}
	//	    }
	//	}
	//	addKeyNum = (addKeyNum)+(span);
	//     }


	//     public static int check()
	//     {
	//	return system("./wbcheck z");
	//     }

	}
