/*  "segs.scm" WB-tree File Based Associative String Data Base System. */
/*  Copyright (C) 1991, 1992, 1993, 2000, 2007 Free Software Foundation, Inc. */
/*  */
/*  This program is free software: you can redistribute it and/or modify */
/*  it under the terms of the GNU Lesser General Public License as */
/*  published by the Free Software Foundation, either version 3 of the */
/*  License, or (at your option) any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, but */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/*  Lesser General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU Lesser General Public */
/*  License along with this program.  If not, see */
/*  <http://www.gnu.org/licenses/>. */

/*  routines in this file return success codes */


#include "wbsys.h"

/* The @var{bsiz} of a segment (given in call to @code{make_seg}) */
/* is a parameter crucial for performance; balancing CPU time traversing */
/* blocks with file-system latency.  @var{bsiz} should be an */
/* integer multiple of the file-system block size. */
/* In the 1990s our nominal @var{bsiz} was 2.kiB; now it should */
/* probably be 4.kiB, 8.kiB, or 16.kiB. */
/*  Segment not in use? */

int seg_free_P(seg)
     int seg;
{
  if (!(lck_tab))
    init_wb(0x4b, 0x96, 0x800);
  if ((0 > (seg))
      || ((seg)>=(num_segs)))
    {
      dprintf((diagout, ">>>>ERROR<<<< bad segment number %d\n", seg));
      return 0;
    }
  else if ((!(seg_port(seg)))
      && (!(seg_str(seg)))
      && (!(seg_used(seg))))
    return !0;
  else return 0;
}


int check_seg_P(seg, filename)
     int seg;
     unsigned char *filename;
{
  int file = blkio_open_read_only_file(filename, 0x80);
  unsigned char tblk[0x19];
  subbytes_read(tblk, 0, 0x19, file);
  {
    int allok_P = (!(str2long(tblk, 0)))
    && (!(str2long(tblk, 4)))
    && (!(str2long(tblk, 8)))
    && (0x77==((((unsigned char*)(tblk))[0x17])))
    && (0x62==((((unsigned char*)(tblk))[0x18])));
    blkio_file_close(file);
    return allok_P;
  }
}


int errout(reason_name, seg, filename)
     unsigned char *reason_name;
     int seg;
     unsigned char *filename;
{
  dprintf((diagout, ">>>>ERROR<<<< not a database %s %s\n", filename, reason_name));
  blkio_file_close(seg_port(seg));
  do_seg_buffers(seg, recycle_ent);
  seg_set_port(seg, 0);
  seg_set_str(seg, 0);
  seg_set_used(seg, 0);
  return typerr;
}

/* TBD - need to lck seg here. */
/*  Segment will be read-only if MODE is #f. */

/*  */
/* @var{seg} should be a non-negative number.  Opens the database @var{seg} in @var{filename} and */
/* returns @var{seg} if successful, a status code otherwise.  The database @var{seg} */
/* will be read-only if the @var{mode} argument is 0.  It will be read-write if */
/* the @var{mode} argument is 2. */

int open_seg(seg, filename, mode)
     int seg;
     unsigned char *filename;
     int mode;
{
  int bsiz = 3*0x80;
  if (!(mode))
    mode = 0;
  if (!(seg_free_P(seg)))
    {
      dprintf((diagout, ">>>>ERROR<<<< %s: segment in use %d\n", "open_seg", seg));
      return argerr;
    }
  else if ((bsiz)>(blk_size))
    {
      dprintf((diagout, ">>>>ERROR<<<< unsupported %s %d > %d\n", "bsiz", bsiz, blk_size));
      return argerr;
    }
  else if (!(check_seg_P(seg, filename)))
    return  -1;
  else {
    int file = mode
      ?blkio_open_modify_file(filename, bsiz)
      :blkio_open_read_only_file(filename, bsiz);
L_loop:
    if (mode
	  ?output_port_P(file)
	  :input_port_P(file))
      {
	seg_set_port(seg, file);
	seg_set_str(seg, filename);
	seg_set_used(seg, 2);
	seg_set_bsiz(seg, bsiz);
	seg_set_flc_len(seg, mode
	    ? -1
	    : -2);
	{
	  HAND *han = seg_rt_han(seg);
	  unsigned char tmp_str[5];
	  if (err_P(bt_open(seg, 0, han, (wcb_sap)+(wcb_sar))))
	    return errout("bt-open 0", seg, filename);
	  else if (2!=(bt_get(han, bsiz_byts, 4, tmp_str)))
	    return errout("BSIZ", seg, filename);
	  else if ((bsiz)!=(str2short(tmp_str, 0)))
	    {
	      blkio_file_close(file);
	      bsiz = str2short(tmp_str, 0);
	      if ((bsiz)>(blk_size))
		return errout("BSIZ too big.", seg, filename);
	      else {
		file = mode
		  ?blkio_open_modify_file(filename, bsiz)
		  :blkio_open_read_only_file(filename, bsiz);
		goto L_loop;
	      }
	    }
	  else if (4!=(bt_get(han, used_byts, 4, tmp_str)))
	    return errout("USED", seg, filename);
	  else {
	    seg_set_used(seg, str2long(tmp_str, 0));
	    if (5!=(bt_get(han, fld_byts, 3, tmp_str)))
	      return errout("FLD", seg, filename);
	    else if (err_P(bt_open(seg, str2long(tmp_str, 1), seg_fl_han(seg), wcb_sar)))
	      return errout("FLC", seg, filename);
	    else {
	      if ((han_typ(seg_fl_han(seg)))!=(frl_typ))
		dprintf((diagout, "Older type freelist - still supported.\n"));
	      han_set_wcb(seg_fl_han(seg), wcb_sar);
	      return seg;
	    }
	  }
	}
      }
    else {
      if (mode
	    ?input_port_P(file)
	    :output_port_P(file))
	blkio_file_close(file);
      dprintf((diagout, ">>>>ERROR<<<< could not open file %s\n", filename));
      return ioerr;
    }
  }
}

/*  */
/* Closes database segment @var{seg} and the file containing it.  If @var{hammer_P} is */
/* NULL, then if there are any problems freeing buffers, then the close */
/* is aborted.  A status code is returned. */

int close_seg(seg, hammer_P)
     int seg;
     int hammer_P;
{
  if ((!(seg_str(seg)))
      || (!(seg_used(seg))))
    {
      dprintf((diagout,
	 ">>>>ERROR<<<< %s: segment %d already closed\n",
	 "close_seg",
	 seg));
      return argerr;
    }
  else {
    flush_flc(seg, 5);
    /*leave only enough blocks to fit in flc in superblock.*/
    if ((seg_flc_len(seg))>=0)
      {
	unsigned char tmp_str[0x14];
	{
	  int i =  -1+(seg_flc_len(seg));
	  while (!(0 > (i))) {
	    long2str(tmp_str, 4*(i), seg_flc(seg)[i]);
	    {
	      i =  -1+(i);
	    }
	  }
	}
	bt_put(seg_rt_han(seg), flc_byts, 3, tmp_str, 4*(seg_flc_len(seg)));
	seg_set_flc_len(seg,  -1);
      }
    {
      int ans = do_seg_buffers(seg, flush_buffer);
      if ((success_P(ans))
	  || (hammer_P))
	{
	  if (!(success_P(ans)))
	    ans = notpres;
	  do_seg_buffers(seg, purge_buffer);
	  bt_close(seg_rt_han(seg));
	  bt_close(seg_fl_han(seg));
	  blkio_file_close(seg_port(seg));
	  seg_set_port(seg, 0);
	  seg_set_str(seg, 0);
	  seg_set_used(seg, 0);
	}
      return ans;
    }
  }
}


unsigned char db_version_str[] = "wb-2a2";

unsigned char db_authors_str[] = "A. Jaffer, J. Finger, R. Zito-Wolf";

/*  */
/*  */
/* The integer @var{seg} must specify an unused segment.  The integer @var{bsiz} */
/* specifies the size of B-tree blocks.  @var{bsiz} should be an integer */
/* multiple of the file-system block size.  Nominal value is 4096. */
/*  */
/* @code{make_seg} makes a new empty database named @var{filename}.  A status code is returned. */
/* Call @code{open_seg} to use the new database. */

int make_seg(seg, filename, bsiz)
     int seg;
     unsigned char *filename;
     int bsiz;
{
  if ((!(seg_free_P(seg)))
      || (!(try_lck_P(seg_lck(seg)))))
    {
      dprintf((diagout, ">>>>ERROR<<<< %s: segment in use %d\n", "make_seg", seg));
      return argerr;
    }
  else if ((bsiz)>(blk_size))
    {
      dprintf((diagout, ">>>>ERROR<<<< unsupported %s %d > %d\n", "bsiz", bsiz, blk_size));
      unlck(seg_lck(seg));
      return argerr;
    }
  else {
    int file = blkio_create_file(filename, bsiz);
    if (output_port_P(file))
      {
	seg_set_port(seg, file);
	seg_set_bsiz(seg, bsiz);
	seg_set_used(seg, 3);
	seg_set_str(seg, filename);
	init_leaf_blk(empty_blk, 0, dir_typ);
	blk_set_time(empty_blk, time(0L));
	blkio_write(file, empty_blk, bsiz, 0);
	init_leaf_blk(empty_blk, 1, dir_typ);
	blk_set_time(empty_blk, time(0L));
	blkio_write(file, empty_blk, bsiz, 1);
	init_leaf_blk(empty_blk, 2, frl_typ);
	blk_set_time(empty_blk, time(0L));
	blkio_write(file, empty_blk, bsiz, 2);
	blkio_file_close(file);
	file = blkio_open_modify_file(filename, bsiz);
	if (output_port_P(file))
	  {
	    seg_set_port(seg, file);
	    unlck(seg_lck(seg));
	    {
	      HAND *han = seg_rt_han(seg);
	      unsigned char tmp_str[5];
	      bt_open(seg, 0, han, (wcb_sap)+(wcb_sar));
	      bt_put(han, no_byts, 0, db_version_str, strlen(db_version_str));
	      long2str(tmp_str, 0, seg_used(seg));
	      bt_put(han, used_byts, 4, tmp_str, 4);
	      short2str(tmp_str, 0, seg_bsiz(seg));
	      bt_put(han, bsiz_byts, 4, tmp_str, 2);
	      tmp_str[0] = 4;
	      long2str(tmp_str, 1, 1);
	      bt_put(han, root_byts, 4, tmp_str, 5);
	      long2str(tmp_str, 1, 2);
	      bt_put(han, fld_byts, 3, tmp_str, 5);
	      bt_put(han, flc_byts, 3, no_byts, 0);
	      if ((bsiz)>0x80)
		bt_put(han, authors_byts, 7, db_authors_str, strlen(db_authors_str));
	      return close_seg(seg, !(!0));
	    }
	  }
	else {
	  dprintf((diagout,
	 ">>>>ERROR<<<< couldn't open fresh file %s\n",
	 filename));
	  unlck(seg_lck(seg));
	  return ioerr;
	}
      }
    else {
      dprintf((diagout,
	 ">>>>ERROR<<<< couldn't create new file %s\n",
	 filename));
      unlck(seg_lck(seg));
      return ioerr;
    }
  }
}

/* @noindent The write-control-bits argument (@var{wcb}) to these */
/* functions controls the latency of updates to the file after various */
/* operations.  These bits are defined as follows: */
/*  */
/* @multitable @columnfractions .1 .3 .6 */
/* @item @var{value} */
/* @tab C-name */
/* @tab Meaning */
/* @item 1 */
/* @tab wcb_sap */
/* @tab save block after PUTs */
/* @cindex wcb_sap */
/* @item 2 */
/* @tab wcb_sar */
/* @tab save block after REMOVEs */
/* @cindex wcb_sar */
/* @item 4 */
/* @tab wcb_sac */
/* @cindex wcb_sac */
/* @tab force block save after cached block changes */
/* @item 8 */
/* @tab wcb_fac */
/* @tab flush buffer entirely after cached block changes (not currently */
/* implemented) */
/* @cindex wcb_fac */
/* @end multitable */
/* Opens bt-handle @var{han} to seg number @var{seg}, block number @var{blk_num}, and */
/* returns the type of the block.  If no such block exists or is not a */
/* root block, then a (negative) status code is returned. */

int bt_open(seg, blk_num, han, wcb)
     int seg;
     long blk_num;
     HAND *han;
     int wcb;
{
  if (((seg)>=0)
      && ((seg)<(num_segs))
      && (seg_str(seg)))
    {
      ENTRY *ent = get_ent(seg, blk_num, accread);
      if (!(ent))
	return unkerr;
      else if (!(root_P(ent_blk(ent))))
	{
	  release_ent(ent, accread);
	  dprintf((diagout, ">>>>ERROR<<<< %s: not a root %d:%ld\n", "bt_open", seg, blk_num));
	  return argerr;
	}
      else {
	han_set_seg(han, seg);
	han_set_num(han, blk_num);
	han_set_typ(han, blk_typ(ent_blk(ent)));
	han_set_last(han, blk_num);
	if (blk_typ_P(ent_blk(ent), dir_typ))
	  wcb = (wcb)|((wcb_sap)+(wcb_sar));
	han_set_wcb(han, wcb);
	release_ent(ent, accread);
	return han_typ(han);
      }
    }
  else return argerr;
}

/* Creates a new root block in seg @var{seg} of type @var{typ}, opens bt-handle */
/* @var{han} to it, and returns a status code.  If @var{seg} has insufficient room to */
/* create a new tree, then the @var{noroom} status code is returned. */
/*  */
/* @code{bt_create} can be used to create temporary b-trees.  Temporary trees will be */
/* be reclaimed by check program after system crashes.  In order to */
/* make a tree persistent, add it to a directory (tree). */

int bt_create(seg, typ, han, wcb)
     int seg;
     int typ;
     HAND *han;
     int wcb;
{
  ENTRY *ent = create_new_blk_ent(seg);
  if (!(ent))
    return noroom;
  else {
    long blk_num = ent_id(ent);
    init_leaf_blk(ent_blk(ent), blk_num, typ);
    ent_set_dty(ent, !0);
    ent_set_pus(ent, 0);
    ents_ent_write(ent);
    han_set_seg(han, seg);
    han_set_num(han, blk_num);
    han_set_typ(han, typ);
    han_set_last(han, blk_num);
    if ((typ)==(dir_typ))
      wcb = (wcb)|((wcb_sap)+(wcb_sar));
    han_set_wcb(han, wcb);
    release_ent(ent, accwrite);
    return success;
  }
}


/* Closes bt-handle @var{han} and returns @var{SUCCESS}. */
/*  */
/* Currently, @code{bt_close} has no effect other than to clear @var{han}. */

int bt_close(han)
     HAND *han;
{
  han_set_seg(han, 0);
  han_set_num(han, 0);
  han_set_typ(han, 0);
  han_set_last(han, 0);
  return success;
}
